from typing import Any, Callable, Iterable, List, Optional, Tuple, Union
import warnings

import numpy as np
import pytorch_lightning
from scipy.sparse import csr_matrix
import torch
from torchmetrics import Metric
from torchmetrics.functional import auroc
from tqdm.auto import tqdm

import collie
from collie.interactions import ExplicitInteractions, Interactions, InteractionsDataLoader
from collie.model import BasePipeline


def _get_user_item_pairs(user_ids: (np.array, torch.tensor),
                         n_items: int,
                         device: Union[str, torch.device]) -> Tuple[torch.tensor, torch.tensor]:
    """
    Create tensors pairing each input user ID with each item ID.

    Parameters
    ----------
    user_ids: np.array or torch.tensor, 1-d
        Iterable[int] of users to score
    n_items: int
        Number of items in the training data
    device: string
        Device to store tensors on

    Returns
    -------
    users: torch.tensor, 1-d
        Tensor with ``n_items`` copies of each user ID
    items: torch.tensor, 1-d
        Tensor with ``len(user_ids)`` copies of each item ID

    Example
    -------
    .. code-block:: python

        >>> user_ids = np.array([10, 11, 12])
        >>> n_items = 4
        >>> user, item = _get_user_item_pairs(user_ids: user_ids, n_items: 4, device: 'cpu'):
        >>> user
        np.array([10, 10, 10, 10, 11, 11, 11, 11, 12, 12, 12, 12])
        >>> item
        np.array([0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3])

    """
    # Added because sometimes we call this function with ``n_items`` as ``np.int64`` type which
    # breaks ``repeat_interleave``.
    if isinstance(n_items, np.int64):
        n_items = n_items.item()

    users = torch.tensor(
        user_ids,
        dtype=torch.int64,
        requires_grad=False,
        device=device,
    ).repeat_interleave(n_items)

    items = torch.arange(
        start=0,
        end=n_items,
        requires_grad=False,
        device=device,
    ).repeat(len(user_ids))

    return users, items


def get_preds(model: BasePipeline,
              user_ids: (np.array, torch.tensor),
              n_items: int,
              device: Union[str, torch.device]) -> torch.tensor:
    """
    Returns a ``n_users x n_items`` tensor with the item IDs of recommended products for each user
    ID.

    Parameters
    ----------
    model: collie.model.BasePipeline
        Model that can take a (user_id, item_id) pair as input and return a recommendation score
    user_ids: np.array or torch.tensor
        Iterable[int] of users to score
    n_items: int
        Number of items in the training data
    device: string
        Device torch should use

    Returns
    -------
    predicted_scores: torch.tensor
        Tensor of shape ``n_users x n_items``

    """
    user, item = _get_user_item_pairs(user_ids, n_items, device)

    with torch.no_grad():
        predicted_scores = model(user, item)

    return predicted_scores.view(-1, n_items)


def _get_labels(targets: csr_matrix,
                user_ids: (np.array, torch.tensor),
                preds: (np.array, torch.tensor),
                device: str) -> torch.tensor:
    """
    Returns a binary array indicating which of the recommended products are in each user's target
    set.

    Parameters
    ----------
    targets: scipy.sparse.csr_matrix
        Interaction matrix containing user and item IDs
    user_ids: np.array or torch.tensor
        Users corresponding to the recommendations in the top k predictions
    preds: torch.tensor
        Top ``k`` item IDs to recommend to each user of shape (n_users x k)
    device: string
        Device torch should use

    Returns
    -------
    labels: torch.tensor
        Tensor with the same dimensions as input ``preds``

    """
    return torch.tensor(
        (targets[user_ids[:, None], np.array(preds.detach().cpu())] > 0)
        .astype('double')
        .toarray(),
        requires_grad=False,
        device=device,
    )


def mapk(targets: csr_matrix,
         user_ids: (np.array, torch.tensor),
         preds: (np.array, torch.tensor),
         k: int = 10) -> float:
    """
    Calculate the mean average precision at K (MAP@K) score for each user.

    Parameters
    ----------
    targets: scipy.sparse.csr_matrix
        Interaction matrix containing user and item IDs
    user_ids: np.array or torch.tensor
        Users corresponding to the recommendations in the top k predictions
    preds: torch.tensor
        Tensor of shape (n_users x n_items) with each user's scores for each item
    k: int
        Number of recommendations to consider per user

    Returns
    -------
    mapk_score: float

    """
    device = preds.device
    n_users = preds.shape[0]

    try:
        predicted_items = preds.topk(k, dim=1).indices
    except RuntimeError as e:
        raise ValueError(
            f'Ensure ``k`` ({k}) is less than the number of items ({preds.shape[1]}):', str(e)
        )

    topk_labeled = _get_labels(targets, user_ids, predicted_items, device)
    accuracy = topk_labeled.int()

    weights = (
        1.0 / torch.arange(
            start=1,
            end=k+1,
            dtype=torch.float64,
            requires_grad=False,
            device=device
        )
    ).repeat(n_users, 1)

    denominator = torch.min(
        torch.tensor(k, device=device, dtype=torch.int).repeat(len(user_ids)),
        torch.tensor(targets[user_ids].getnnz(axis=1), device=device)
    )

    res = ((accuracy * accuracy.cumsum(axis=1) * weights).sum(axis=1)) / denominator
    res[torch.isnan(res)] = 0

    return res.mean().item()


def mrr(targets: csr_matrix,
        user_ids: (np.array, torch.tensor),
        preds: (np.array, torch.tensor),
        k: Optional[Any] = None) -> float:
    """
    Calculate the mean reciprocal rank (MRR) of the input predictions.

    Parameters
    ----------
    targets: scipy.sparse.csr_matrix
        Interaction matrix containing user and item IDs
    user_ids: np.array or torch.tensor
        Users corresponding to the recommendations in the top k predictions
    preds: torch.tensor
        Tensor of shape (n_users x n_items) with each user's scores for each item
    k: Any
        Ignored, included only for compatibility with ``mapk``

    Returns
    -------
    mrr_score: float

    """
    predicted_items = preds.topk(preds.shape[1], dim=1).indices
    labeled = _get_labels(targets, user_ids, predicted_items, device=preds.device)

    # weighting each 0/1 by position so that topk returns index of *first* postive result
    position_weight = 1.0/(
        torch.arange(1, targets.shape[1] + 1, device=preds.device)
        .repeat(len(user_ids), 1)
        .float()
    )
    labeled_weighted = (labeled.float() * position_weight)

    highest_score, rank = labeled_weighted.topk(k=1)

    reciprocal_rank = 1.0/(rank.float() + 1)
    reciprocal_rank[highest_score == 0] = 0

    return reciprocal_rank.mean().item()


def auc(targets: csr_matrix,
        user_ids: (np.array, torch.tensor),
        preds: (np.array, torch.tensor),
        k: Optional[Any] = None) -> float:
    """
    Calculate the area under the ROC curve (AUC) for each user and average the results.

    Parameters
    ----------
    targets: scipy.sparse.csr_matrix
        Interaction matrix containing user and item IDs
    user_ids: np.array or torch.tensor
        Users corresponding to the recommendations in the top k predictions
    preds: torch.tensor
        Tensor of shape (n_users x n_items) with each user's scores for each item
    k: Any
        Ignored, included only for compatibility with ``mapk``

    Returns
    -------
    auc_score: float

    """
    agg = 0
    for i, user_id in enumerate(user_ids):
        target_tensor = torch.tensor(
            targets[user_id].toarray(),
            device=preds.device,
            dtype=torch.long
        ).view(-1)
        # many models' ``preds`` may be unbounded if a final activation layer is not applied
        # we have to normalize ``preds`` here to avoid a ``ValueError`` stating that ``preds``
        # should be probabilities, but values were detected outside of [0,1] range
        auc = auroc(torch.sigmoid(preds[i, :]), target=target_tensor, pos_label=1)
        agg += auc

    return (agg/len(user_ids)).item()


def evaluate_in_batches(
    metric_list: Iterable[Callable],
    test_interactions: collie.interactions.Interactions,
    model: collie.model.BasePipeline,
    k: int = 10,
    batch_size: int = 20,
    logger: pytorch_lightning.loggers.base.LightningLoggerBase = None,
    verbose: bool = True,
) -> List[float]:
    """
    Evaluate a model with potentially several different metrics.

    Memory constraints require that most test sets will need to be evaluated in batches. This
    function handles the looping and batching boilerplate needed to properly evaluate the model
    without running out of memory.

    Parameters
    ----------
    metric_list: list of functions
        List of evaluation functions to apply. Each function must accept keyword arguments:

        * ``targets``

        * ``user_ids``

        * ``preds``

        * ``k``

    test_interactions: collie.interactions.Interactions
        Interactions to use as labels
    model: collie.model.BasePipeline
        Model that can take a (user_id, item_id) pair as input and return a recommendation score
    k: int
        Number of recommendations to consider per user. This is ignored by some metrics
    batch_size: int
        Number of users to score in a single batch. For best efficiency, this number should be as
        high as possible without running out of memory
    logger: pytorch_lightning.loggers.base.LightningLoggerBase
        If provided, will log outputted metrics dictionary using the ``log_metrics`` method with
        keys being the string representation of ``metric_list`` and values being
        ``evaluation_results``. Additionally, if ``model.hparams.num_epochs_completed`` exists, this
        will be logged as well, making it possible to track metrics progress over the course of
        model training
    verbose: bool
        Display progress bar and print statements during function execution

    Returns
    -------
    evaluation_results: list
        List of floats, with each metric value corresponding to the respective function passed in
        ``metric_list``

    Examples
    --------
    .. code-block:: python

        from collie.metrics import auc, evaluate_in_batches, mapk, mrr


        map_10_score, mrr_score, auc_score = evaluate_in_batches(
            metric_list=[mapk, mrr, auc],
            test_interactions=test,
            model=model,
        )

        print(map_10_score, mrr_score, auc_score)

    """
    if not isinstance(test_interactions, Interactions):
        raise ValueError(
            '``test_interactions`` must be of type ``Interactions``, not '
            f'{type(test_interactions)}. Try using ``explicit_evaluate_in_batches`` instead.'
        )

    device = _get_evaluate_in_batches_device(model=model)
    model.to(device)

    test_users = np.unique(test_interactions.mat.row)
    targets = test_interactions.mat.tocsr()

    if len(test_users) < batch_size:
        batch_size = len(test_users)

    accumulators = [0] * len(metric_list)

    data_to_iterate_over = range(int(np.ceil(len(test_users) / batch_size)))
    if verbose:
        data_to_iterate_over = tqdm(data_to_iterate_over)

    for i in data_to_iterate_over:
        user_range = test_users[i * batch_size:(i + 1) * batch_size]
        preds = get_preds(model, user_range, test_interactions.num_items, device)
        for metric_ind, metric in enumerate(metric_list):
            score = metric(targets=targets, user_ids=user_range, preds=preds, k=k)
            accumulators[metric_ind] += (score * len(user_range))

    all_scores = [acc_score / len(test_users) for acc_score in accumulators]

    if logger is not None:
        _log_metrics(model=model,
                     logger=logger,
                     metric_list=metric_list,
                     all_scores=all_scores,
                     verbose=verbose)

    return all_scores[0] if len(all_scores) == 1 else all_scores


def explicit_evaluate_in_batches(
    metric_list: Iterable[Metric],
    test_interactions: collie.interactions.ExplicitInteractions,
    model: collie.model.BasePipeline,
    logger: pytorch_lightning.loggers.base.LightningLoggerBase = None,
    verbose: bool = True,
    **kwargs,
) -> List[float]:
    """
    Evaluate a model with potentially several different metrics.

    Memory constraints require that most test sets will need to be evaluated in batches. This
    function handles the looping and batching boilerplate needed to properly evaluate the model
    without running out of memory.

    Parameters
    ----------
    metric_list: list of ``torchmetrics.Metric``
        List of evaluation functions to apply. Each function must accept arguments for predictions
        and targets, in order
    test_interactions: collie.interactions.ExplicitInteractions
    model: collie.model.BasePipeline
        Model that can take a (user_id, item_id) pair as input and return a recommendation score
    batch_size: int
        Number of users to score in a single batch. For best efficiency, this number should be as
        high as possible without running out of memory
    logger: pytorch_lightning.loggers.base.LightningLoggerBase
        If provided, will log outputted metrics dictionary using the ``log_metrics`` method with
        keys being the string representation of ``metric_list`` and values being
        ``evaluation_results``. Additionally, if ``model.hparams.num_epochs_completed`` exists, this
        will be logged as well, making it possible to track metrics progress over the course of
        model training
    verbose: bool
        Display progress bar and print statements during function execution
    kwargs: keyword arguments
        Additional arguments sent to the ``InteractionsDataLoader``

    Returns
    ----------
    evaluation_results: list
        List of floats, with each metric value corresponding to the respective function passed in
        ``metric_list``

    Examples
    -------------
    .. code-block:: python

        import torchmetrics

        from collie.metrics import explicit_evaluate_in_batches


        mse_score, mae_score = evaluate_in_batches(
            metric_list=[torchmetrics.MeanSquaredError(), torchmetrics.MeanAbsoluteError()],
            test_interactions=test,
            model=model,
        )

        print(mse_score, mae_score)

    """
    if not isinstance(test_interactions, ExplicitInteractions):
        raise ValueError(
            '``test_interactions`` must be of type ``ExplicitInteractions``, not '
            f'{type(test_interactions)}. Try using ``evaluate_in_batches`` instead.'
        )

    try:
        device = _get_evaluate_in_batches_device(model=model)
        model.to(device)

        test_loader = InteractionsDataLoader(interactions=test_interactions,
                                             **kwargs)

        data_to_iterate_over = test_loader
        if verbose:
            data_to_iterate_over = tqdm(test_loader)

        for batch in data_to_iterate_over:
            users, items, ratings = batch

            # move data to batch before sending to model
            users = users.to(device)
            items = items.to(device)
            ratings = ratings.cpu()

            preds = model(users, items)

            for metric in metric_list:
                metric(preds.cpu(), ratings)

        all_scores = [metric.compute() for metric in metric_list]

        if logger is not None:
            _log_metrics(model=model,
                         logger=logger,
                         metric_list=metric_list,
                         all_scores=all_scores,
                         verbose=verbose)

        return all_scores[0] if len(all_scores) == 1 else all_scores
    finally:
        for metric in metric_list:
            metric.reset()


def _get_evaluate_in_batches_device(model: BasePipeline):
    device = getattr(model, 'device') or ('cuda:0' if torch.cuda.is_available() else 'cpu')

    if torch.cuda.is_available() and getattr(model, 'device') == 'cpu':
        warnings.warn('CUDA available but model device is set to CPU - is this desired?')

    return device


def _log_metrics(model: BasePipeline,
                 logger: pytorch_lightning.loggers.base.LightningLoggerBase,
                 metric_list: List[Union[Callable, Metric]],
                 all_scores: List[float],
                 verbose: bool):
    try:
        step = model.hparams.get('num_epochs_completed')
    except torch.nn.modules.module.ModuleAttributeError:
        # if, somehow, there is no ``model.hparams`` attribute, this shouldn't fail
        step = None

    try:
        metrics_dict = dict(zip([x.__name__ for x in metric_list], all_scores))
    except AttributeError:
        metrics_dict = dict(zip([type(x).__name__ for x in metric_list], all_scores))

    if verbose:
        print(f'Logging metrics {metrics_dict} to ``logger``...')

    logger.log_metrics(metrics=metrics_dict, step=step)



import pytest
import torch

from collie.interactions import ExplicitInteractions, Interactions


@pytest.fixture()
def test_implicit_interactions():
    return Interactions(
        users=[0, 0, 0, 1, 1, 1, 2, 2],
        items=[0, 1, 2, 1, 2, 3, 0, 2],
        ratings=[1, 1, 1, 1, 1, 1, 1, 1],
        check_num_negative_samples_is_valid=False,
    )


@pytest.fixture()
def test_implicit_predicted_scores():
    return torch.tensor([
        [0.9, 0.4, 0.5, 0.7],
        [0.1, 1.2, 1.7, 0.12],
        [-1, -2, -3, 4],
    ])


@pytest.fixture()
def test_explicit_interactions():
    return ExplicitInteractions(
        users=[0, 0, 0, 1, 1, 1, 2, 2],
        items=[0, 1, 2, 1, 2, 3, 0, 2],
        ratings=[1, 2, 3, 4, 5, 4, 3, 2],
    )


@pytest.fixture()
def test_explicit_predicted_scores():
    return torch.tensor([
        [0.9, 1.4, 2.5, -0.7],
        [5.1, 4.2, 3.7, 2.12],
        [-1, 2, -3, 4],
    ])


@pytest.fixture()
def test_implicit_recs():
    return torch.tensor([
        [0, 3, 2, 1],
        [2, 1, 3, 0],
        [3, 0, 1, 2],
    ])


@pytest.fixture()
def test_implicit_labels():
    return torch.tensor([
        [1, 0, 1, 1],
        [1, 1, 1, 0],
        [0, 1, 0, 1],
    ], dtype=float)


@pytest.fixture()
def targets(test_implicit_interactions):
    return test_implicit_interactions.mat.tocsr()


@pytest.fixture()
def metrics():
    return {
        'mapk': 0.7685185,
        'mrr': 0.8333333,
        'auc': 0.5277777,
        'mse': 5.86055,
        'mae': 1.69750,
    }



from collections import defaultdict
import functools
import operator
from typing import Any, Iterable, Optional, Tuple, Union

from joblib import delayed, Parallel
import numpy as np
from scipy.sparse import coo_matrix
from sklearn.model_selection import train_test_split

from collie.interactions import (BaseInteractions,
                                 ExplicitInteractions,
                                 HDF5Interactions,
                                 Interactions)
from collie.utils import get_random_seed


def _subset_interactions(interactions: BaseInteractions,
                         idxs: Iterable[int]) -> Union[ExplicitInteractions, Interactions]:
    idxs = np.array(idxs)

    coo_mat = coo_matrix(
        (interactions.mat.data[idxs], (interactions.mat.row[idxs], interactions.mat.col[idxs])),
        shape=(interactions.num_users, interactions.num_items)
    )

    # disable all ``Interactions`` checks for the data splits, since we assume the initial
    # ``Interactions`` object would have these checks already applied prior to the data split
    if isinstance(interactions, Interactions):
        return Interactions(
            mat=coo_mat,
            num_negative_samples=interactions.num_negative_samples,
            allow_missing_ids=True,
            remove_duplicate_user_item_pairs=False,
            num_users=interactions.num_users,
            num_items=interactions.num_items,
            check_num_negative_samples_is_valid=False,
            max_number_of_samples_to_consider=interactions.max_number_of_samples_to_consider,
            seed=interactions.seed,
        )
    else:
        return ExplicitInteractions(
            mat=coo_mat,
            allow_missing_ids=True,
            remove_duplicate_user_item_pairs=False,
            num_users=interactions.num_users,
            num_items=interactions.num_items,
        )


def random_split(interactions: BaseInteractions,
                 val_p: float = 0.0,
                 test_p: float = 0.2,
                 processes: Optional[Any] = None,
                 seed: Optional[int] = None) -> Tuple[BaseInteractions, ...]:
    """
    Randomly split interactions into training, validation, and testing sets.

    This split does NOT guarantee that every user will be represented in both the training and
    testing datasets. While much faster than ``stratified_split``, it is not the most representative
    data split because of this.

    Note that this function is not supported for ``HDF5Interactions`` objects, since this data split
    implementation requires all data to fit in memory. A data split for large datasets should be
    done using a big data processing technology, like Spark.

    Parameters
    ----------
    interactions: collie.interactions.BaseInteractions
    val_p: float
        Proportion of data used for validation
    test_p: float
        Proportion of data used for testing
    processes: Any
        Ignored, included only for compatability with ``stratified_split`` API
    seed: int
        Random seed for splits

    Returns
    -------
    train_interactions: collie.interactions.BaseInteractions
        Training data of size proportional to ``1 - val_p - test_p``
    validate_interactions: collie.interactions.BaseInteractions
        Validation data of size proportional to ``val_p``, returned only if ``val_p > 0``
    test_interactions: collie.interactions.BaseInteractions
        Testing data of size proportional to ``test_p``

    Examples
    --------
    .. code-block:: python

        >>> interactions = Interactions(...)
        >>> len(interactions)
        100000
        >>> train, test = random_split(interactions)
        >>> len(train), len(test)
        (80000, 20000)

    """
    assert not isinstance(interactions, HDF5Interactions), (
        '``HDF5Interactions`` data type not supported in cross validation splits!'
    )

    _validate_val_p_and_test_p(val_p=val_p, test_p=test_p)

    if seed is None:
        seed = get_random_seed()

    np.random.seed(seed)

    shuffle_indices = np.arange(len(interactions))
    np.random.shuffle(shuffle_indices)

    interactions = _subset_interactions(interactions=interactions,
                                        idxs=shuffle_indices)

    validate_and_test_p = val_p + test_p
    validate_cutoff = int((1.0 - validate_and_test_p) * len(interactions))
    test_cutoff = int((1.0 - test_p) * len(interactions))

    train_idxs = np.arange(validate_cutoff)
    validate_idxs = np.arange(validate_cutoff, test_cutoff)
    test_idxs = np.arange(test_cutoff, len(interactions))

    train_interactions = _subset_interactions(interactions=interactions,
                                              idxs=train_idxs)
    test_interactions = _subset_interactions(interactions=interactions,
                                             idxs=test_idxs)

    if val_p > 0:
        validate_interactions = _subset_interactions(interactions=interactions,
                                                     idxs=validate_idxs)

        return train_interactions, validate_interactions, test_interactions
    else:
        return train_interactions, test_interactions


def stratified_split(interactions: BaseInteractions,
                     val_p: float = 0.0,
                     test_p: float = 0.2,
                     processes: int = -1,
                     seed: Optional[int] = None) -> Tuple[BaseInteractions, ...]:
    """
    Split an ``Interactions`` instance into train, validate, and test datasets in a stratified
    manner such that each user appears at least once in each of the datasets.

    This split guarantees that every user will be represented in the training, validation, and
    testing datasets given they appear in ``interactions`` at least three times. If ``val_p ==
    0``, they will appear in the training and testing datasets given they appear at least two times.
    If a user appears fewer than this number of times, a ``ValueError`` will
    be raised. To filter users with fewer than ``n`` points out, use
    ``collie.utils.remove_users_with_fewer_than_n_interactions``.

    This is computationally more complex than ``random_split``, but produces a more representative
    data split. Note that when ``val_p > 0``, the algorithm will perform the data split twice,
    once to create the test set and another to create the validation set, essentially doubling the
    computational time.

    Note that this function is not supported for ``HDF5Interactions`` objects, since this data split
    implementation requires all data to fit in memory. A data split for large datasets should be
    done using a big data processing technology, like Spark.

    Parameters
    ----------
    interactions: collie.interactions.BaseInteractions
        ``Interactions`` instance containing the data to split
    val_p: float
        Proportion of data used for validation
    test_p: float
        Proportion of data used for testing
    processes: int
        Number of CPUs to use for parallelization. If ``processes == 0``, this will be run
        sequentially in a single list comprehension, else this function uses ``joblib.delayed`` and
        ``joblib.Parallel`` for parallelization. A value of ``-1`` means that all available cores
        will be used
    seed: int
        Random seed for splits

    Returns
    -------
    train_interactions: collie.interactions.BaseInteractions
        Training data of size proportional to ``1 - val_p - test_p``
    validate_interactions: collie.interactions.BaseInteractions
        Validation data of size proportional to ``val_p``, returned only if ``val_p > 0``
    test_interactions: collie.interactions.BaseInteractions
        Testing data of size proportional to ``test_p``

    Examples
    --------
    .. code-block:: python

        >>> interactions = Interactions(...)
        >>> len(interactions)
        100000
        >>> train, test = stratified_split(interactions)
        >>> len(train), len(test)
        (80000, 20000)

    """
    assert not isinstance(interactions, HDF5Interactions), (
        '``HDF5Interactions`` data types not supported in cross validation splits!'
    )

    _validate_val_p_and_test_p(val_p=val_p, test_p=test_p)

    if seed is None:
        seed = get_random_seed()

    train, test = _stratified_split(interactions=interactions,
                                    test_p=test_p,
                                    processes=processes,
                                    seed=seed)

    if val_p > 0:
        train, validate = _stratified_split(interactions=train,
                                            test_p=val_p / (1 - test_p),
                                            processes=processes,
                                            seed=seed)

        return train, validate, test
    else:
        return train, test


def _stratified_split(interactions: BaseInteractions,
                      test_p: float,
                      processes: int,
                      seed: int) -> Tuple[Interactions, Interactions]:
    users = interactions.mat.row
    unique_users = set(users)

    # while we should be able to run ``np.where(users == user)[0]`` to find all items each user
    # interacted with, by building up a dictionary to get these values instead, we can achieve the
    # same result in O(N) complexity rather than O(M * N), a nice timesave to have when working with
    # larger datasets
    all_idxs_for_users_dict = defaultdict(list)
    for idx, user in enumerate(users):
        all_idxs_for_users_dict[user].append(idx)

    if processes == 0:
        test_idxs = [
            _stratified_split_parallel_worker(idxs_to_split=all_idxs_for_users_dict[user],
                                              test_p=test_p,
                                              seed=(seed + user))
            for user in unique_users
        ]
    else:
        # run the function below in parallel for each user
        # by setting the seed to ``seed + user``, we get a balance between reproducability and
        # actual randomness so users with the same number of interactions are not split the exact
        # same way
        test_idxs = Parallel(n_jobs=processes)(
            delayed(_stratified_split_parallel_worker)(all_idxs_for_users_dict[user],
                                                       test_p,
                                                       seed + user)
            for user in unique_users
        )

    # reduce the list of lists down to a 1-d list
    test_idxs = functools.reduce(operator.iconcat, test_idxs, [])
    # find all indices not in test set - they are now train
    train_idxs = list(set(range(len(users))) - set(test_idxs))

    train_interactions = _subset_interactions(interactions=interactions,
                                              idxs=train_idxs)
    test_interactions = _subset_interactions(interactions=interactions,
                                             idxs=test_idxs)

    return train_interactions, test_interactions


def _stratified_split_parallel_worker(idxs_to_split: Iterable[Any],
                                      test_p: float, seed: int) -> np.array:
    _, test_idxs = train_test_split(idxs_to_split,
                                    test_size=test_p,
                                    random_state=seed,
                                    shuffle=True,
                                    stratify=np.ones_like(idxs_to_split))

    return test_idxs


def _validate_val_p_and_test_p(val_p: float, test_p: float) -> None:
    validate_and_test_p = val_p + test_p

    if val_p >= 1 or val_p < 0:
        raise ValueError('``val_p`` must be in the range [0, 1).')
    if test_p >= 1 or test_p < 0:
        raise ValueError('``test_p`` must be in the range [0, 1).')
    if validate_and_test_p >= 1 or validate_and_test_p <= 0:
        raise ValueError('The sum of ``val_p`` and ``test_p`` must be in the range (0, 1).')



from typing import Dict, Optional

import torch


def ideal_difference_from_metadata(
    positive_items: torch.tensor,
    negative_items: torch.tensor,
    metadata: Optional[Dict[str, torch.tensor]],
    metadata_weights: Optional[Dict[str, float]],
) -> torch.tensor:
    """
    Helper function to calculate the ideal score difference between the positive and negative items.

    Without considering metadata, the ideal score difference would be 1.0 (since the function looks
    at a pair of items, one positive item and one negative item). Taking metadata into
    consideration, the ideal score difference should be between 0 and 1 if there is a partial match
    (not the same item, but matching metadata - e.g. the same film genre). This function calculates
    that ideal difference when there is metadata available.

    Metadata passed in to this function is independent of metadata given to the model during
    training - it can be the same data or a different set. For example, one might use genre
    embeddings as metadata during training and use genre labels as metadata during loss calculation
    (since all metadata passed in to this function must be categorical).

    Parameters
    ----------
    positive_items: torch.tensor, 1-d
        Tensor containing IDs for known positive items
    negative_items: torch.tensor, 1-d
        Tensor containing IDs for sampled negative items
    metadata: dict
        Keys should be strings identifying each metadata type that match keys in
        ``metadata_weights``. Values should be a ``torch.tensor`` of shape (num_items x 1). Each
        tensor should contain categorical metadata information about items (e.g. a number
        representing the genre of the item)
    metadata_weights: dict
        Keys should be strings identifying each metadata type that match keys in ``metadata``.
        Values should be the amount of weight to place on a match of that type of metadata, with the
        sum of all values ``<= 1``.
        e.g. If ``metadata_weights = {'genre': .3, 'director': .2}``, then an item is:

        * a 100% match if it's the same item,

        * a 50% match if it's a different item with the same genre and same director,

        * a 30% match if it's a different item with the same genre and different director,

        * a 20% match if it's a different item with a different genre and same director,

        * a 0% match if it's a different item with a different genre and different director,
          which is equivalent to the loss without any partial credit

    Returns
    -------
    ideal difference: torch.tensor
        Tensor with the same shape as ``positive_items``, with each element between 0 and 1

    """
    weight_sum = sum(metadata_weights.values())
    if weight_sum > 1:
        raise ValueError(f'sum of metadata weights was {weight_sum}, must be <=1')

    match_frac = torch.zeros(positive_items.shape).to(positive_items.device)
    for k, array in metadata.items():
        array = array.squeeze()
        match_frac += (
            array[positive_items.long()] == array[negative_items.long()]
        ).int().to(positive_items.device)*metadata_weights[k]

    return 1.0 - match_frac



import os
from pathlib import Path
import re
import zipfile

import pandas as pd
import requests

from collie.config import DATA_PATH


def read_movielens_df(decrement_ids: bool = True) -> pd.DataFrame:
    """
    Read ``u.data`` from the MovieLens 100K dataset.

    If there is not a directory at ``$DATA_PATH/ml-100k``, this function creates that directory and
    downloads the entire dataset there.

    See the MovieLens 100K README for additional information on the dataset:
    https://files.grouplens.org/datasets/movielens/ml-100k-README.txt

    Parameters
    ----------
    decrement_ids: bool
        Decrement user and item IDs by 1 before returning, which is required for Collie's
        ``Interactions`` dataset

    Returns
    -------
    df: pd.DataFrame
        MovieLens 100K ``u.data`` comprising of columns:

            * user_id

            * item_id

            * rating

            * timestamp

    Side Effects
    ------------
    Creates directory at ``$DATA_PATH/ml-100k`` and downloads data files if data does not exist.

    """
    _make_data_path_dirs_if_not_exist()

    df_path = os.path.join(DATA_PATH, 'ml-100k', 'u.data')
    if not Path(df_path).exists():
        _download_movielens_100k()

    column_names = ['user_id', 'item_id', 'rating', 'timestamp']
    df = pd.read_csv(df_path, sep='\t', names=column_names)

    if decrement_ids:
        df.loc[:, 'user_id'] = df['user_id'] - 1
        df.loc[:, 'item_id'] = df['item_id'] - 1

    return df


def read_movielens_df_item() -> pd.DataFrame:
    """
    Read ``u.item`` from the MovieLens 100K dataset.

    If there is not a directory at ``$DATA_PATH/ml-100k``, this function creates that directory and
    downloads the entire dataset there.

    See the MovieLens 100K README for additional information on the dataset:
    https://files.grouplens.org/datasets/movielens/ml-100k-README.txt

    Returns
    -------
    df_item: pd.DataFrame
        MovieLens 100K ``u.item`` containing columns:

            * item_id

            * movie_title

            * release_date

            * video_release_date

            * IMDb_URL

            * unknown

            * Action

            * Adventure

            * Animation

            * Children

            * Comedy

            * Crime

            * Documentary

            * Drama

            * Fantasy

            * Film_Noir

            * Horror

            * Musical

            * Mystery

            * Romance', 'Sci_Fi

            * Thriller

            * War

            * Wester

    Side Effects
    ------------
    Creates directory at ``$DATA_PATH/ml-100k`` and downloads data files if data does not exist.

    """
    _make_data_path_dirs_if_not_exist()

    df_item_path = os.path.join(DATA_PATH, 'ml-100k', 'u.item')
    if not Path(df_item_path).exists():
        _download_movielens_100k()

    column_names = ['item_id', 'movie_title', 'release_date', 'video_release_date', 'IMDb_URL',
                    'unknown', 'Action', 'Adventure', 'Animation', 'Children', 'Comedy', 'Crime',
                    'Documentary', 'Drama', 'Fantasy', 'Film_Noir', 'Horror', 'Musical', 'Mystery',
                    'Romance', 'Sci_Fi', 'Thriller', 'War', 'Western']
    df_item = pd.read_csv(df_item_path, sep='|', encoding='latin-1', names=column_names)

    df_item['release_date'] = pd.to_datetime(df_item['release_date'])
    df_item = df_item.drop(columns=['video_release_date'])

    return df_item


def _make_data_path_dirs_if_not_exist() -> None:
    """Get path to the movielens dataset file."""
    if not DATA_PATH.exists():
        print(f'Making data path at ``{DATA_PATH}``...')
        DATA_PATH.mkdir(parents=True, exist_ok=True)


def _download_movielens_100k() -> None:
    """Download the MovieLens 100K data."""
    url = 'http://files.grouplens.org/datasets/movielens/ml-100k.zip'
    req = requests.get(url, stream=True)

    print('Downloading MovieLens 100K data...')

    with open(os.path.join(DATA_PATH, 'ml-100k.zip'), 'wb') as f:
        f.write(req.content)

    with zipfile.ZipFile(os.path.join(DATA_PATH / 'ml-100k.zip'), 'r') as z:
        z.extractall(DATA_PATH)


def read_movielens_posters_df() -> pd.DataFrame:
    """
    Read in data containing the item ID and poster URL for visualization purposes of MovieLens 100K
    data.

    This function will attempt to read the file at ``data/movielens_posters.csv`` if it exists and,
    if not, will read the CSV from the origin GitHub repo at
    https://raw.githubusercontent.com/ShopRunner/collie/main/data/movielens_posters.csv.

    Returns
    -------
    posters_df: pd.DataFrame
        DataFrame comprising columns:

            * item_id

            * url

    """
    # attempt to first load from a local file
    absolute_data_path = Path(__file__).parent.absolute().parent.parent / 'data'
    movielens_posters_csv_filepath = os.path.join(absolute_data_path, 'movielens_posters.csv')

    # be prepared to read the CSV from the origin GitHub repo as well
    movielens_posters_csv_url = (
        'https://raw.githubusercontent.com/ShopRunner/collie/main/data/movielens_posters.csv'
    )

    posters_df = pd.read_csv(
        movielens_posters_csv_filepath
        if os.path.exists(movielens_posters_csv_filepath)
        else movielens_posters_csv_url
    )

    return posters_df


def get_movielens_metadata(df_item: pd.DataFrame = None) -> pd.DataFrame:
    """
    Return MovieLens 100K metadata as a DataFrame.

    DataFrame returned has the following column order:

    .. code-block:: python

        [
            'genre_action', 'genre_adventure', 'genre_animation', 'genre_children', 'genre_comedy',
            'genre_crime', 'genre_documentary', 'genre_drama', 'genre_fantasy', 'genre_film_noir',
            'genre_horror', 'genre_musical', 'genre_mystery', 'genre_romance', 'genre_sci_fi',
            'genre_thriller', 'genre_war', 'genre_western', 'genre_unknown', 'decade_unknown',
            'decade_20', 'decade_30', 'decade_40', 'decade_50', 'decade_60',
            'decade_70', 'decade_80', 'decade_90',
        ]

    See the MovieLens 100K README for additional information on the dataset:
    https://files.grouplens.org/datasets/movielens/ml-100k-README.txt

    Parameters
    ----------
    df_item: pd.DataFrame
        DataFrame of MovieLens 100K ``u.item`` containing binary columns of movie names and
        metadata. If ``None``, will automatically read the output of ``read_movielens_df_item()``

    Returns
    -------
    metadata_df: pd.DataFrame

    """
    if df_item is None:
        df_item = read_movielens_df_item()

    # format movies decade
    df_item_date = df_item.iloc[:, [2]].copy()
    df_item_date.loc[:, 'year'] = df_item_date['release_date'].dt.year.fillna(1900)
    df_item_date.loc[:, 'decade'] = ((df_item_date['year'] - 1900) / 10).astype('int64') * 10
    df_decades = pd.get_dummies(df_item_date.decade, prefix='decade')
    df_decades.columns = ['decade_unknown'] + df_decades.columns[1:].tolist()

    # format movie genre
    df_item_genre = df_item.iloc[:, list(range(4, 23))].copy()
    df_item_genre.columns = 'genre_' + df_item_genre.columns.str.lower()

    # format final metadata structure
    metadata_df = pd.merge(df_item_genre, df_decades, left_index=True, right_index=True)

    # find and swap genre_unknown to end of genre list
    cols = metadata_df.columns.values.tolist()
    last_genre_element = list(filter(re.compile('genre*').match, cols))[-1]
    last_genre_index = cols.index(last_genre_element)
    cols.insert(last_genre_index + 1, 'genre_unknown')
    cols.remove('genre_unknown')
    metadata_df = metadata_df[cols]

    return metadata_df



from functools import partial
import os
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

import joblib
import numpy as np
import pandas as pd
import torch
from torch import nn
import torch.nn.functional as F
from torch.optim.lr_scheduler import ReduceLROnPlateau

from collie.config import DATA_PATH
from collie.interactions import (ApproximateNegativeSamplingInteractionsDataLoader,
                                 Interactions,
                                 InteractionsDataLoader)
from collie.model.base import MultiStagePipeline, ScaledEmbedding, ZeroEmbedding
from collie.utils import get_init_arguments, merge_docstrings


INTERACTIONS_LIKE_INPUT = Union[ApproximateNegativeSamplingInteractionsDataLoader,
                                Interactions,
                                InteractionsDataLoader]


class HybridModel(MultiStagePipeline):
    # NOTE: the full docstring is merged in with ``MultiStagePipeline``'s using
    # ``merge_docstrings``. Only the description of new or changed parameters are included in this
    # docstring
    """
    Training pipeline for a multi-stage hybrid recommendation model.

    ``HybridModel`` models contain dense layers that process item metadata, concatenate this
    embedding with user and item embeddings, sending this concatenated embedding through more dense
    layers to output a single float ranking / rating. We add both user and item biases to this score
    before returning. This is the same architecture as the ``HybridPretrainedModel``, but we are
    training the embeddings ourselves rather than relying on pulling this from a pre-trained model.

    The stages in a ``HybridModel`` are, in order:

    1. ``matrix_factorization``
        Matrix factorization exactly as we do in ``MatrixFactorizationModel``. In this stage,
        metadata is NOT incorporated into the model.
    2. ``metadata_only``
        User and item embeddings terms are frozen, and the MLP layers for the metadata (if
        specified) and combined embedding-metadata data are optimized.
    3. ``all``
        Embedding and MLP layers are all optimized together, including those for metadata.

    All ``HybridModel`` instances are subclasses of the ``LightningModule`` class provided by
    PyTorch Lightning. This means to train a model, you will need a
    ``collie.model.CollieTrainer`` object, but the model can be saved and loaded without this
    ``Trainer`` instance. Example usage may look like:

    .. code-block:: python

        from collie.model import CollieTrainer, HybridModel


        # instantiate and fit a ``HybridModel`` as expected
        model = HybridModel(train=train, item_metadata=item_metadata)
        trainer = CollieTrainer(model)
        trainer.fit(model)

        # train for X more epochs on the next stage, ``metadata_only``
        trainer.max_epochs += X
        model.advance_stage()
        trainer.fit(model)

        # train for Y more epochs on the next stage, ``all``
        trainer.max_epochs += Y
        model.advance_stage()
        trainer.fit(model)

        model.eval()

        # do evaluation as normal with ``model``

        model.save_model(path='model')
        new_model = HybridModel(load_model_path='model')

        # do evaluation as normal with ``new_model``

    Note
    ----
    The ``forward`` calculation will be different depending on the stage that is set. Note this
    when evaluating / saving and loading models in.

    Parameters
    ----------
    item_metadata: torch.tensor, pd.DataFrame, or np.array, 2-dimensional
        The shape of the item metadata should be (num_items x metadata_features), and each item's
        metadata should be available when indexing a row by an item ID
    embedding_dim: int
        Number of latent factors to use for user and item embeddings
    metadata_layers_dims: list
        List of linear layer dimensions to apply to the metadata only, starting with the dimension
        directly following ``metadata_features`` and ending with the dimension to concatenate with
        the item embeddings
    combined_layers_dims: list
        List of linear layer dimensions to apply to the concatenated item embeddings and item
        metadata, starting with the dimension directly following the shape of
        ``item_embeddings + metadata_features`` and ending with the dimension before the final
        linear layer to dimension 1
    dropout_p: float
        Probability of dropout
    metadata_only_stage_lr: float
        Learning rate for metadata and combined layers optimized during the ``metadata_only`` stage
    all_stage_lr: float
        Learning rate for all model parameters optimized during the ``all`` stage
    optimizer: torch.optim or str
        Optimizer used for embeddings and bias terms (if ``bias_optimizer`` is ``None``) during the
        ``matrix_factorization`` stage. If a string, one of the following supported optimizers:

        * ``'sgd'`` (for ``torch.optim.SGD``)

        * ``'adam'`` (for ``torch.optim.Adam``)

    metadata_only_stage_optimizer: torch.optim or str
        Optimizer used for metadata and combined layers during the ``metadata_only`` stage. If a
        string, one of the following supported optimizers:

        * ``'sgd'`` (for ``torch.optim.SGD``)

        * ``'adam'`` (for ``torch.optim.Adam``)

    all_stage_optimizer: torch.optim or str
        Optimizer used for all model parameters during the ``all`` stage. If a string, one of the
        following supported optimizers:

        * ``'sgd'`` (for ``torch.optim.SGD``)

        * ``'adam'`` (for ``torch.optim.Adam``)

    """
    def __init__(self,
                 train: INTERACTIONS_LIKE_INPUT = None,
                 val: INTERACTIONS_LIKE_INPUT = None,
                 item_metadata: Union[torch.tensor, pd.DataFrame, np.array] = None,
                 embedding_dim: int = 30,
                 metadata_layers_dims: Optional[List[int]] = None,
                 combined_layers_dims: List[int] = [128, 64, 32],
                 dropout_p: float = 0.0,
                 lr: float = 1e-3,
                 bias_lr: Optional[Union[float, str]] = 1e-2,
                 metadata_only_stage_lr: float = 1e-3,
                 all_stage_lr: float = 1e-4,
                 lr_scheduler_func: Optional[Callable] = partial(
                     ReduceLROnPlateau,
                     patience=1,
                     verbose=False,
                 ),
                 weight_decay: float = 0.0,
                 optimizer: Union[str, Callable] = 'adam',
                 bias_optimizer: Optional[Union[str, Callable]] = 'sgd',
                 metadata_only_stage_optimizer: Union[str, Callable] = 'adam',
                 all_stage_optimizer: Union[str, Callable] = 'adam',
                 loss: Union[str, Callable] = 'hinge',
                 metadata_for_loss: Optional[Dict[str, torch.tensor]] = None,
                 metadata_for_loss_weights: Optional[Dict[str, float]] = None,
                 load_model_path: Optional[str] = None,
                 map_location: Optional[str] = None):
        item_metadata_num_cols = None
        optimizer_config_list = None

        if load_model_path is None:
            if item_metadata is None:
                raise ValueError('Must provide item metadata for ``HybridPretrainedModel``.')
            elif isinstance(item_metadata, pd.DataFrame):
                item_metadata = torch.from_numpy(item_metadata.to_numpy())
            elif isinstance(item_metadata, np.ndarray):
                item_metadata = torch.from_numpy(item_metadata)

            item_metadata = item_metadata.float()

            item_metadata_num_cols = item_metadata.shape[1]

            if bias_optimizer is not None:
                initial_optimizer_block = [
                    {
                        'lr': lr,
                        'optimizer': optimizer,
                        # optimize embeddings...
                        'parameter_prefix_list': ['user_embedding', 'item_embedding'],
                        'stage': 'matrix_factorization',
                    },
                    {
                        'lr': lr if bias_lr == 'infer' else bias_lr,
                        'optimizer': optimizer if bias_optimizer == 'infer' else bias_optimizer,
                        # ... and optimize bias terms too
                        'parameter_prefix_list': ['user_bias', 'item_bias'],
                        'stage': 'matrix_factorization',
                    },
                ]
            else:
                initial_optimizer_block = [
                    {
                        'lr': lr,
                        'optimizer': optimizer,
                        # optimize embeddings and bias terms all together
                        'parameter_prefix_list': [
                            'user_embedding',
                            'item_embedding',
                            'user_bias',
                            'item_bias'],
                        'stage': 'matrix_factorization',
                    },
                ]

            optimizer_config_list = initial_optimizer_block + [
                {
                    'lr': metadata_only_stage_lr,
                    'optimizer': metadata_only_stage_optimizer,
                    # optimize metadata layers only
                    'parameter_prefix_list': ['metadata', 'combined', 'user_bias', 'item_bias'],
                    'stage': 'metadata_only',
                },
                {
                    'lr': all_stage_lr,
                    'optimizer': all_stage_optimizer,
                    # optimize everything
                    'parameter_prefix_list': ['user', 'item', 'metadata', 'combined'],
                    'stage': 'all',
                },
            ]

        super().__init__(optimizer_config_list=optimizer_config_list,
                         item_metadata_num_cols=item_metadata_num_cols,
                         **get_init_arguments())

    __doc__ = merge_docstrings(MultiStagePipeline, __doc__, __init__)

    def _move_any_external_data_to_device(self):
        """Move item metadata to the device before training."""
        self.item_metadata = self.item_metadata.to(self.device)

    def _load_model_init_helper(self, load_model_path: str, map_location: str, **kwargs) -> None:
        self.item_metadata = (
            joblib.load(os.path.join(load_model_path, 'metadata.pkl'))
        )
        super()._load_model_init_helper(load_model_path=os.path.join(load_model_path, 'model.pth'),
                                        map_location=map_location,
                                        **kwargs)

    def _setup_model(self, **kwargs) -> None:
        """
        Method for building model internals that rely on the data passed in.

        This method will be called after `prepare_data`.

        """
        if self.hparams.load_model_path is None:
            if not hasattr(self, 'item_metadata'):
                self.item_metadata = kwargs.pop('item_metadata')

        self.user_biases = ZeroEmbedding(num_embeddings=self.hparams.num_users,
                                         embedding_dim=1)
        self.item_biases = ZeroEmbedding(num_embeddings=self.hparams.num_items,
                                         embedding_dim=1)
        self.user_embeddings = ScaledEmbedding(num_embeddings=self.hparams.num_users,
                                               embedding_dim=self.hparams.embedding_dim)
        self.item_embeddings = ScaledEmbedding(num_embeddings=self.hparams.num_items,
                                               embedding_dim=self.hparams.embedding_dim)
        self.dropout = nn.Dropout(p=self.hparams.dropout_p)

        # set up metadata-only layers
        metadata_output_dim = self.hparams.item_metadata_num_cols
        self.metadata_layers = None
        if self.hparams.metadata_layers_dims is not None:
            metadata_layers_dims = (
                [self.hparams.item_metadata_num_cols] + self.hparams.metadata_layers_dims
            )
            self.metadata_layers = [
                nn.Linear(metadata_layers_dims[idx - 1], metadata_layers_dims[idx])
                for idx in range(1, len(metadata_layers_dims))
            ]
            for i, layer in enumerate(self.metadata_layers):
                nn.init.xavier_normal_(self.metadata_layers[i].weight)
                self.add_module('metadata_layer_{}'.format(i), layer)

            metadata_output_dim = metadata_layers_dims[-1]

        # set up combined layers
        combined_dimension_input = (
            self.user_embeddings.embedding_dim
            + self.item_embeddings.embedding_dim
            + metadata_output_dim
        )
        combined_layers_dims = [combined_dimension_input] + self.hparams.combined_layers_dims + [1]
        self.combined_layers = [
            nn.Linear(combined_layers_dims[idx - 1], combined_layers_dims[idx])
            for idx in range(1, len(combined_layers_dims))
        ]
        for i, layer in enumerate(self.combined_layers):
            nn.init.xavier_normal_(self.combined_layers[i].weight)
            self.add_module('combined_layer_{}'.format(i), layer)

    def forward(self, users: torch.tensor, items: torch.tensor) -> torch.tensor:
        """
        Forward pass through the model.

        Parameters
        ----------
        users: tensor, 1-d
            Array of user indices
        items: tensor, 1-d
            Array of item indices

        Returns
        -------
        preds: tensor, 1-d
            Predicted ratings or rankings

        """
        if self.hparams.stage == 'matrix_factorization':
            pred_scores = (
                torch.mul(
                    self.dropout(self.user_embeddings(users)),
                    self.dropout(self.item_embeddings(items))
                ).sum(axis=1)
                + self.user_biases(users).squeeze(1)
                + self.item_biases(items).squeeze(1)
            )
        else:
            # TODO: remove self.device and let lightning do it
            metadata_output = self.item_metadata[items, :].to(self.device)
            if self.metadata_layers is not None:
                for metadata_nn_layer in self.metadata_layers:
                    metadata_output = self.dropout(
                        F.leaky_relu(
                            metadata_nn_layer(metadata_output)
                        )
                    )

            # TODO: make this matrix factorization instead of only a MLP
            combined_output = torch.cat((self.user_embeddings(users),
                                         self.item_embeddings(items),
                                         metadata_output), 1)
            for combined_nn_layer in self.combined_layers[:-1]:
                combined_output = self.dropout(
                    F.leaky_relu(
                        combined_nn_layer(combined_output)
                    )
                )

            pred_scores = (
                self.combined_layers[-1](combined_output)
                + self.user_biases(users)
                + self.item_biases(items)
            )

        return pred_scores.squeeze()

    def _get_item_embeddings(self) -> torch.tensor:
        """Get item embeddings on device."""
        # TODO: update this to get the embeddings post-MLP
        return self.item_embeddings.weight.data

    def save_model(self,
                   path: Union[str, Path] = os.path.join(DATA_PATH / 'model'),
                   overwrite: bool = False) -> None:
        """
        Save the model's state dictionary, hyperparameters, and item metadata.

        While PyTorch Lightning offers a way to save and load models, there are two main reasons
        for overriding these:

        1) To properly save and load a model requires the ``Trainer`` object, meaning that all
           deployed models will require Lightning to run the model, which is not actually needed
           for inference.

        2) In the v0.8.4 release, loading a model back in leads to a ``RuntimeError`` unable to load
           in weights.

        Parameters
        ----------
        path: str or Path
            Directory path to save model and data files
        overwrite: bool
            Whether or not to overwrite existing data

        """
        path = str(path)

        if os.path.exists(path):
            if os.listdir(path) and overwrite is False:
                raise ValueError(f'Data exists in ``path`` at {path} and ``overwrite`` is False.')

        Path(path).mkdir(parents=True, exist_ok=True)
        joblib.dump(self.item_metadata, os.path.join(path, 'metadata.pkl'))

        super().save_model(filename=os.path.join(path, 'model.pth'))



from functools import partial
from typing import Callable, Dict, Iterable, Optional, Union

import pandas as pd
import torch
from torch import nn
from torch.optim.lr_scheduler import ReduceLROnPlateau

from collie.interactions import (ApproximateNegativeSamplingInteractionsDataLoader,
                                 Interactions,
                                 InteractionsDataLoader)
from collie.model import MultiStagePipeline, ScaledEmbedding, ZeroEmbedding
from collie.utils import get_init_arguments, merge_docstrings


INTERACTIONS_LIKE_INPUT = Union[ApproximateNegativeSamplingInteractionsDataLoader,
                                Interactions,
                                InteractionsDataLoader]


class ColdStartModel(MultiStagePipeline):
    # NOTE: the full docstring is merged in with ``MultiStagePipeline``'s using
    # ``merge_docstrings``. Only the description of new or changed parameters are included in this
    # docstring
    """
    Training pipeline for a matrix factorization model optimized for the cold-start problem.

    Many recommendation models suffer from the cold start problem, when a model is unable to provide
    adequate recommendations for a new item until enough users have interacted with it. But, if
    users only interact with recommended items, the item will never be recommended, and thus the
    model will never improve recommendations for this item.

    The ``ColdStartModel`` attempts to bypass this by limiting the item space down to "item
    buckets", training a model on this as the item space, then expanding out to all items. During
    this expansion, the learned-embeddings of each bucket is copied over to each corresponding
    item, providing a smarter initialization than a random one for both existing and new items.
    Now, when we have a new item, we can use its bucket embedding as an initialization into a model.

    The stages in a ``ColdStartModel`` are, in order:

    1. ``item_buckets``
        Matrix factorization with item embeddings and bias terms bucketed by
        ``item_buckets`` argument. Unlike in the next stage, many items may map on to a single
        bucket, and this will share the same embedding and bias representation. The model should
        learn user preference for buckets in this stage.
    2. ``no_buckets``
        Standard matrix factorization as we do in ``MatrixFactorizationModel``. However, upon
        advancing to this stage, the item embeddings are initialized with their bucketed embedding
        value (and same for biases). Not only does this provide better initialization than random,
        but allows new items to be incorporated into the model without training by using their
        item bucket embedding and bias terms at prediction time.

    Note that the cold start problem exists for new users as well, but this functionality will be
    added to this model in a future version.

    All ``ColdStartModel`` instances are subclasses of the ``LightningModule`` class provided by
    PyTorch Lightning. This means to train a model, you will need a
    ``collie.model.CollieTrainer`` object, but the model can be saved and loaded without this
    ``Trainer`` instance. Example usage may look like:

    .. code-block:: python

        from collie.model import ColdStartModel, CollieTrainer


        # instantiate and fit a ``ColdStartModel`` as expected
        model = ColdStartModel(train=train, item_buckets=item_buckets)
        trainer = CollieTrainer(model)
        trainer.fit(model)

        # train for X more epochs on the next stage, ``no_buckets``
        trainer.max_epochs += X
        model.advance_stage()
        trainer.fit(model)

        model.eval()

        # do evaluation as normal with ``model``

        # get item-item recommendations for a new item by using the bucket ID, Z
        similar_items = model.item_bucket_item_similarity(item_bucket_id=Z)

        model.save_model(filename='model.pth')
        new_model = ColdStartModel(load_model_path='model.pth')

        # do evaluation as normal with ``new_model``

    Note
    ----
    The ``forward`` calculation will be different depending on the stage that is set. Note this
    when evaluating / saving and loading models in.

    Parameters
    ----------
    item_buckets: torch.tensor, 1-d
        An ordered iterable containing the bucket ID for each item ID. For example, if you have
        five films and are going to bucket by primary genre, and your data looks like:

        * Item ID: 0, Genre ID: 1

        * Item ID: 1, Genre ID: 0

        * Item ID: 2, Genre ID: 2

        * Item ID: 3, Genre ID: 2

        * Item ID: 4, Genre ID: 1

        Then ``item_buckets`` would be: ``[1, 0, 2, 2, 1]``
    embedding_dim: int
        Number of latent factors to use for user and item embeddings
    dropout_p: float
        Probability of dropout
    item_buckets_stage_lr: float
        Learning rate for user parameters and item bucket parameters optimized during the
        ``item_buckets`` stage
    no_buckets_stage_lr: float
        Learning rate for user parameters and item parameters optimized during the ``no_buckets``
        stage
    item_buckets_stage_lr: float
        Optimizer used for user parameters and item bucket parameters optimized during the
        ``item_buckets`` stage. If a string, one of the following supported optimizers:

        * ``'sgd'`` (for ``torch.optim.SGD``)

        * ``'adam'`` (for ``torch.optim.Adam``)

    no_buckets_stage_lr: float
        Optimizer used for user parameters and item parameters optimized during the ``no_buckets``
        stage. If a string, one of the following supported optimizers:

        * ``'sgd'`` (for ``torch.optim.SGD``)

        * ``'adam'`` (for ``torch.optim.Adam``)

    """
    def __init__(self,
                 train: INTERACTIONS_LIKE_INPUT = None,
                 val: INTERACTIONS_LIKE_INPUT = None,
                 item_buckets: Iterable[int] = None,
                 embedding_dim: int = 30,
                 dropout_p: float = 0.0,
                 sparse: bool = False,
                 item_buckets_stage_lr: float = 1e-3,
                 no_buckets_stage_lr: float = 1e-3,
                 lr_scheduler_func: Optional[Callable] = partial(
                     ReduceLROnPlateau,
                     patience=1,
                     verbose=False,
                 ),
                 weight_decay: float = 0.0,
                 item_buckets_stage_optimizer: Union[str, Callable] = 'adam',
                 no_buckets_stage_optimizer: Union[str, Callable] = 'adam',
                 loss: Union[str, Callable] = 'hinge',
                 metadata_for_loss: Optional[Dict[str, torch.tensor]] = None,
                 metadata_for_loss_weights: Optional[Dict[str, float]] = None,
                 load_model_path: Optional[str] = None,
                 map_location: Optional[str] = None):
        optimizer_config_list = None
        num_item_buckets = None

        if load_model_path is None:
            # TODO: separate out optimizer and bias optimizer somehow
            optimizer_config_list = [
                {
                    'lr': item_buckets_stage_lr,
                    'optimizer': item_buckets_stage_optimizer,
                    'parameter_prefix_list': [
                        'user_embed',
                        'user_bias',
                        'item_bucket_embed',
                        'item_bucket_bias',
                    ],
                    'stage': 'item_buckets',
                },
                {
                    'lr': no_buckets_stage_lr,
                    'optimizer': no_buckets_stage_optimizer,
                    'parameter_prefix_list': [
                        'user_embed',
                        'user_bias',
                        'item_embed',
                        'item_bias',
                    ],
                    'stage': 'no_buckets',
                },
            ]

            if not isinstance(item_buckets, torch.Tensor):
                item_buckets = torch.tensor(item_buckets)

            # data quality checks for ``item_buckets``
            assert item_buckets.dim() == 1, (
                f'``item_buckets`` must be 1-dimensional, not {item_buckets.dim()}-dimensional!'
            )
            if len(item_buckets) != train.num_items:
                raise ValueError(
                    'Length of ``item_buckets`` must be equal to the number of items in the '
                    f'dataset: {len(item_buckets)} != {train.num_items}.'
                )
            if min(item_buckets) != 0:
                raise ValueError(f'``item_buckets`` IDs must start at 0, not {min(item_buckets)}!')

            num_item_buckets = item_buckets.max().item() + 1

        super().__init__(optimizer_config_list=optimizer_config_list,
                         num_item_buckets=num_item_buckets,
                         **get_init_arguments())

    __doc__ = merge_docstrings(MultiStagePipeline, __doc__, __init__)

    def _move_any_external_data_to_device(self):
        """Move the item buckets to the device before training."""
        self.hparams.item_buckets = self.hparams.item_buckets.to(self.device)

    def _copy_weights(self, old: nn.Embedding, new: nn.Embedding, buckets: torch.tensor) -> None:
        new.weight.data.copy_(old.weight.data[buckets])

    def set_stage(self, stage: str) -> None:
        """Set the stage for the model."""
        current_stage = self.hparams.stage

        if stage in self.hparams.stage_list:
            if current_stage == 'item_buckets' and stage == 'no_buckets':
                print('Copying over item embeddings...')
                self._copy_weights(self.item_bucket_biases,
                                   self.item_biases,
                                   self.hparams.item_buckets)
                self._copy_weights(self.item_bucket_embeddings,
                                   self.item_embeddings,
                                   self.hparams.item_buckets)
        else:
            raise ValueError(
                f'"{stage}" is not a valid stage, please choose one of {self.hparams.stage_list}'
            )

        self.hparams.stage = stage
        print(f'Set ``self.hparams.stage`` to "{stage}"')

    def _setup_model(self, **kwargs) -> None:
        """
        Method for building model internals that rely on the data passed in.

        This method will be called after `prepare_data`.

        """
        # define initial embedding groups
        self.item_bucket_biases = ZeroEmbedding(
            num_embeddings=self.hparams.num_item_buckets,
            embedding_dim=1,
            sparse=self.hparams.sparse,
        )
        self.item_bucket_embeddings = ScaledEmbedding(
            num_embeddings=self.hparams.num_item_buckets,
            embedding_dim=self.hparams.embedding_dim,
            sparse=self.hparams.sparse,
        )

        # define fine-tuned embedding groups
        self.user_biases = ZeroEmbedding(
            num_embeddings=self.hparams.num_users,
            embedding_dim=1,
            sparse=self.hparams.sparse
        )
        self.item_biases = ZeroEmbedding(
            num_embeddings=self.hparams.num_items,
            embedding_dim=1,
            sparse=self.hparams.sparse,
        )
        self.user_embeddings = ScaledEmbedding(
            num_embeddings=self.hparams.num_users,
            embedding_dim=self.hparams.embedding_dim,
            sparse=self.hparams.sparse
        )
        self.item_embeddings = ScaledEmbedding(
            num_embeddings=self.hparams.num_items,
            embedding_dim=self.hparams.embedding_dim,
            sparse=self.hparams.sparse,
        )

        self.dropout = nn.Dropout(p=self.hparams.dropout_p)

    def forward(self, users: torch.tensor, items: torch.tensor) -> torch.tensor:
        """
        Forward pass through the model.

        Parameters
        ----------
        users: tensor, 1-d
            Array of user indices
        items: tensor, 1-d
            Array of item indices

        Returns
        -------
        preds: tensor, 1-d
            Predicted ratings or rankings

        """
        user_embeddings = self.user_embeddings(users)
        user_biases = self.user_biases(users)

        if self.hparams.stage == 'item_buckets':
            # transform item IDs to item bucket IDs
            items = self.hparams.item_buckets[items]

            item_embeddings = self.item_bucket_embeddings(items)
            item_biases = self.item_bucket_biases(items)
        elif self.hparams.stage == 'no_buckets':
            item_embeddings = self.item_embeddings(items)
            item_biases = self.item_biases(items)

        pred_scores = (
            torch.mul(self.dropout(user_embeddings), self.dropout(item_embeddings)).sum(axis=1)
            + user_biases.squeeze(1)
            + item_biases.squeeze(1)
        )

        return pred_scores.squeeze()

    def item_bucket_item_similarity(self, item_bucket_id: int) -> pd.Series:
        """
        Get most similar item indices to a item bucket by cosine similarity.

        Cosine similarity is computed with item and item bucket embeddings from a trained model.

        Parameters
        ----------
        item_id: int

        Returns
        -------
        sim_score_idxs: pd.Series
            Sorted values as cosine similarity for each item in the dataset with the index being
            the item ID

        """
        item_bucket_embeddings = self.item_bucket_embeddings.weight.data
        item_bucket_embeddings = (
            item_bucket_embeddings / item_bucket_embeddings.norm(dim=1)[:, None]
        )

        item_embeddings = self._get_item_embeddings()
        item_embeddings = item_embeddings / item_embeddings.norm(dim=1)[:, None]

        sim_score_idxs = (
            torch.matmul(item_bucket_embeddings[[item_bucket_id], :],
                         item_embeddings.transpose(1, 0))
            .detach()
            .cpu()
            .numpy()
            .squeeze()
        )

        sim_score_idxs_series = pd.Series(sim_score_idxs)
        sim_score_idxs_series = sim_score_idxs_series.sort_values(ascending=False)

        return sim_score_idxs_series

    def _get_item_embeddings(self) -> torch.tensor:
        """Get item embeddings on device."""
        return self.item_embeddings.weight.data



from ._version import __version__

from collie.config import *
from collie.cross_validation import *
from collie.interactions import *
from collie.loss import *
from collie.metrics import *
from collie.model import *
from collie.movielens import *
from collie.utils import *



from collections import OrderedDict
import copy
from functools import partial
import os
from pathlib import Path
from typing import Callable, Dict, List, Optional, Union

import joblib
import numpy as np
import pandas as pd
import torch
from torch import nn
import torch.nn.functional as F
from torch.optim.lr_scheduler import ReduceLROnPlateau

from collie.config import DATA_PATH
from collie.model.base import BasePipeline, INTERACTIONS_LIKE_INPUT, ScaledEmbedding, ZeroEmbedding
from collie.model.matrix_factorization import MatrixFactorizationModel
from collie.utils import get_init_arguments, merge_docstrings


class HybridPretrainedModel(BasePipeline):
    # NOTE: the full docstring is merged in with ``BasePipeline``'s using ``merge_docstrings``.
    # Only the description of new or changed parameters are included in this docstring
    """
    Training pipeline for a hybrid recommendation model using a pre-trained matrix factorization
    model as its base.

    ``HybridPretrainedModel`` models contain dense layers that process item metadata, concatenate
    this embedding with the user and item embeddings copied from a trained
    ``MatrixFactorizationModel``, and send this concatenated embedding through more dense layers to
    output a single float ranking / rating. We add both user and item biases to this score before
    returning. This is the same architecture as the ``HybridModel``, but we are using the embeddings
    from a pre-trained model rather than training them up ourselves.

    All ``HybridPretrainedModel`` instances are subclasses of the ``LightningModule`` class
    provided by PyTorch Lightning. This means to train a model, you will need a
    ``collie.model.CollieTrainer`` object, but the model can be saved and loaded without this
    ``Trainer`` instance. Example usage may look like:

    .. code-block:: python

        from collie.model import CollieTrainer, HybridPretrainedModel, MatrixFactorizationModel


        # instantiate and fit a ``MatrixFactorizationModel`` as expected
        mf_model = MatrixFactorizationModel(train=train)
        mf_trainer = CollieTrainer(mf_model)
        mf_trainer.fit(mf_model)

        hybrid_model = HybridPretrainedModel(train=train,
                                             item_metadata=item_metadata,
                                             trained_model=mf_model)
        hybrid_trainer = CollieTrainer(hybrid_model)
        hybrid_trainer.fit(hybrid_model)
        hybrid_model.eval()

        # do evaluation as normal with ``hybrid_model``

        hybrid_model.save_model(path='model')
        new_hybrid_model = HybridPretrainedModel(load_model_path='model')

        # do evaluation as normal with ``new_hybrid_model``

    Parameters
    ----------
    item_metadata: torch.tensor, pd.DataFrame, or np.array, 2-dimensional
        The shape of the item metadata should be (num_items x metadata_features), and each item's
        metadata should be available when indexing a row by an item ID
    trained_model: ``collie.model.MatrixFactorizationModel``
        Previously trained ``MatrixFactorizationModel`` model to extract embeddings from
    metadata_layers_dims: list
        List of linear layer dimensions to apply to the metadata only, starting with the dimension
        directly following ``metadata_features`` and ending with the dimension to concatenate with
        the item embeddings
    combined_layers_dims: list
        List of linear layer dimensions to apply to the concatenated item embeddings and item
        metadata, starting with the dimension directly following the shape of
        ``item_embeddings + metadata_features`` and ending with the dimension before the final
        linear layer to dimension 1
    freeze_embeddings: bool
        When initializing the model, whether or not to freeze ``trained_model``'s embeddings
    dropout_p: float
        Probability of dropout
    optimizer: torch.optim or str
        If a string, one of the following supported optimizers:

        * ``'sgd'`` (for ``torch.optim.SGD``)

        * ``'adam'`` (for ``torch.optim.Adam``)

    """
    def __init__(self,
                 train: INTERACTIONS_LIKE_INPUT = None,
                 val: INTERACTIONS_LIKE_INPUT = None,
                 item_metadata: Union[torch.tensor, pd.DataFrame, np.array] = None,
                 trained_model: MatrixFactorizationModel = None,
                 metadata_layers_dims: Optional[List[int]] = None,
                 combined_layers_dims: List[int] = [128, 64, 32],
                 freeze_embeddings: bool = True,
                 dropout_p: float = 0.0,
                 lr: float = 1e-3,
                 lr_scheduler_func: Optional[Callable] = partial(ReduceLROnPlateau,
                                                                 patience=1,
                                                                 verbose=True),
                 weight_decay: float = 0.0,
                 optimizer: Union[str, Callable] = 'adam',
                 loss: Union[str, Callable] = 'hinge',
                 metadata_for_loss: Optional[Dict[str, torch.tensor]] = None,
                 metadata_for_loss_weights: Optional[Dict[str, float]] = None,
                 # y_range: Optional[Tuple[float, float]] = None,
                 load_model_path: Optional[str] = None,
                 map_location: Optional[str] = None):
        item_metadata_num_cols = None
        if load_model_path is None:
            if trained_model is None:
                raise ValueError('Must provide ``trained_model`` for ``HybridPretrainedModel``.')

            if item_metadata is None:
                raise ValueError('Must provide item metadata for ``HybridPretrainedModel``.')
            elif isinstance(item_metadata, pd.DataFrame):
                item_metadata = torch.from_numpy(item_metadata.to_numpy())
            elif isinstance(item_metadata, np.ndarray):
                item_metadata = torch.from_numpy(item_metadata)

            item_metadata = item_metadata.float()

            item_metadata_num_cols = item_metadata.shape[1]

        super().__init__(**get_init_arguments(),
                         item_metadata_num_cols=item_metadata_num_cols)

    __doc__ = merge_docstrings(BasePipeline, __doc__, __init__)

    def _move_any_external_data_to_device(self):
        """Move item metadata to the device before training."""
        self.item_metadata = self.item_metadata.to(self.device)

    def _load_model_init_helper(self, load_model_path: str, map_location: str, **kwargs) -> None:
        self.item_metadata = (
            joblib.load(os.path.join(load_model_path, 'metadata.pkl'))
        )
        super()._load_model_init_helper(load_model_path=os.path.join(load_model_path, 'model.pth'),
                                        map_location=map_location)

    def _setup_model(self, **kwargs) -> None:
        """
        Method for building model internals that rely on the data passed in.

        This method will be called after ``prepare_data``.

        """
        if self.hparams.load_model_path is None:
            if not hasattr(self, '_trained_model'):
                self._trained_model = kwargs.pop('trained_model')
            if not hasattr(self, 'item_metadata'):
                self.item_metadata = kwargs.pop('item_metadata')

            # we are not loading in a model, so we will create a new model from scratch
            # we don't want to modify the ``trained_model``'s weights, so we deep copy
            self.embeddings = nn.Sequential(
                copy.deepcopy(self._trained_model.user_embeddings),
                copy.deepcopy(self._trained_model.item_embeddings)
            )
            self.biases = nn.Sequential(
                copy.deepcopy(self._trained_model.user_biases),
                copy.deepcopy(self._trained_model.item_biases)
            )

            if self.hparams.freeze_embeddings:
                self.freeze_embeddings()
            else:
                self.unfreeze_embeddings()

            # save hyperparameters that we need to be able to rebuilt the embedding layers on load
            self.hparams.user_num_embeddings = self.embeddings[0].num_embeddings
            self.hparams.user_embeddings_dim = self.embeddings[0].embedding_dim
            self.hparams.item_num_embeddings = self.embeddings[1].num_embeddings
            self.hparams.item_embeddings_dim = self.embeddings[1].embedding_dim
        else:
            # assume we are loading in a previously-saved model
            # set up dummy embeddings with the correct dimensions so we can load weights in
            self.embeddings = nn.Sequential(
                ScaledEmbedding(self.hparams.user_num_embeddings, self.hparams.user_embeddings_dim),
                ScaledEmbedding(self.hparams.item_num_embeddings, self.hparams.item_embeddings_dim)
            )
            self.biases = nn.Sequential(
                ZeroEmbedding(self.hparams.user_num_embeddings, 1),
                ZeroEmbedding(self.hparams.item_num_embeddings, 1)
            )

        self.dropout = nn.Dropout(p=self.hparams.dropout_p)

        # set up metadata-only layers
        metadata_output_dim = self.hparams.item_metadata_num_cols
        self.metadata_layers = None
        if self.hparams.metadata_layers_dims is not None:
            metadata_layers_dims = (
                [self.hparams.item_metadata_num_cols] + self.hparams.metadata_layers_dims
            )
            self.metadata_layers = [
                nn.Linear(metadata_layers_dims[idx - 1], metadata_layers_dims[idx])
                for idx in range(1, len(metadata_layers_dims))
            ]
            for i, layer in enumerate(self.metadata_layers):
                nn.init.xavier_normal_(self.metadata_layers[i].weight)
                self.add_module('metadata_layer_{}'.format(i), layer)

            metadata_output_dim = metadata_layers_dims[-1]

        # set up combined layers
        combined_dimension_input = (
            self.hparams.user_embeddings_dim
            + self.hparams.item_embeddings_dim
            + metadata_output_dim
        )
        combined_layers_dims = [combined_dimension_input] + self.hparams.combined_layers_dims + [1]
        self.combined_layers = [
            nn.Linear(combined_layers_dims[idx - 1], combined_layers_dims[idx])
            for idx in range(1, len(combined_layers_dims))
        ]
        for i, layer in enumerate(self.combined_layers):
            nn.init.xavier_normal_(self.combined_layers[i].weight)
            self.add_module('combined_layer_{}'.format(i), layer)

    def forward(self, users: torch.tensor, items: torch.tensor) -> torch.tensor:
        """
        Forward pass through the model.

        Parameters
        ----------
        users: tensor, 1-d
            Array of user indices
        items: tensor, 1-d
            Array of item indices

        Returns
        -------
        preds: tensor, 1-d
            Predicted ratings or rankings

        """
        if str(self.device) != str(self.item_metadata.device):
            self._move_any_external_data_to_device()

        metadata_output = self.item_metadata[items, :]
        if self.metadata_layers is not None:
            for metadata_nn_layer in self.metadata_layers:
                metadata_output = self.dropout(
                    F.leaky_relu(
                        metadata_nn_layer(metadata_output)
                    )
                )

        combined_output = torch.cat((self.embeddings[0](users),
                                     self.embeddings[1](items),
                                     metadata_output), 1)
        for combined_nn_layer in self.combined_layers[:-1]:
            combined_output = self.dropout(
                F.leaky_relu(
                    combined_nn_layer(combined_output)
                )
            )

        pred_scores = (
            self.combined_layers[-1](combined_output)
            + self.biases[0](users)
            + self.biases[1](items)
        )

        return pred_scores.squeeze()

    def _get_item_embeddings(self) -> torch.tensor:
        """Get item embeddings on device."""
        # TODO: update this to get the embeddings post-MLP
        return self.embeddings[1].weight.data

    def freeze_embeddings(self) -> None:
        """Remove gradient requirement from the embeddings."""
        self.embeddings[0].weight.requires_grad = False
        self.embeddings[1].weight.requires_grad = False

    def unfreeze_embeddings(self) -> None:
        """Require gradients for the embeddings."""
        self.embeddings[0].weight.requires_grad = True
        self.embeddings[1].weight.requires_grad = True

    def save_model(self,
                   path: Union[str, Path] = os.path.join(DATA_PATH / 'model'),
                   overwrite: bool = False) -> None:
        """
        Save the model's state dictionary, hyperparameters, and item metadata.

        While PyTorch Lightning offers a way to save and load models, there are two main reasons
        for overriding these:

        1) To properly save and load a model requires the ``Trainer`` object, meaning that all
           deployed models will require Lightning to run the model, which is not actually needed
           for inference.

        2) In the v0.8.4 release, loading a model back in leads to a ``RuntimeError`` unable to load
           in weights.

        Parameters
        ----------
        path: str or Path
            Directory path to save model and data files
        overwrite: bool
            Whether or not to overwrite existing data

        """
        path = str(path)

        if os.path.exists(path):
            if os.listdir(path) and overwrite is False:
                raise ValueError(f'Data exists in ``path`` at {path} and ``overwrite`` is False.')

        Path(path).mkdir(parents=True, exist_ok=True)
        joblib.dump(self.item_metadata, os.path.join(path, 'metadata.pkl'))

        # preserve ordering while extracting the state dictionary without the ``_trained_model``
        # component
        state_dict_keys_to_save = [
            k for k, _ in self.state_dict().items() if '_trained_model' not in k
        ]
        state_dict_vals_to_save = [
            v for k, v in self.state_dict().items() if '_trained_model' not in k
        ]
        state_dict_to_save = OrderedDict(zip(state_dict_keys_to_save, state_dict_vals_to_save))

        dict_to_save = {'state_dict': state_dict_to_save, 'hparams': self.hparams}
        torch.save(dict_to_save, os.path.join(path, 'model.pth'))

    def load_from_hybrid_model(self, hybrid_model) -> None:
        """
        Copy hyperparameters and state dictionary from an existing ``HybridPretrainedModel``
        instance.

        This is particularly useful for creating another PyTorch Lightning trainer object to
        fine-tune copied-over embeddings from a ``MatrixFactorizationModel`` instance.

        Parameters
        ----------
        hybrid_model: ``collie.model.HybridPretrainedModel``
            HybridPretrainedModel containing hyperparameters and state dictionary to copy over

        """
        for key, value in hybrid_model.hparams.items():
            self.hparams[key] = value

        self._setup_model()
        self.load_state_dict(state_dict=hybrid_model.state_dict())
        self.eval()



from typing import Dict, Optional, Union

import torch

from collie.loss.metadata_utils import ideal_difference_from_metadata


def warp_loss(
    positive_scores: torch.tensor,
    many_negative_scores: torch.tensor,
    num_items: int,
    positive_items: Optional[torch.tensor] = None,
    negative_items: Optional[torch.tensor] = None,
    metadata: Optional[Dict[str, torch.tensor]] = dict(),
    metadata_weights: Optional[Dict[str, float]] = dict(),
) -> torch.tensor:
    """
    Modified WARP loss function [4]_.

    See http://www.thespermwhale.com/jaseweston/papers/wsabie-ijcai.pdf for loss equation.

    See ``ideal_difference_from_metadata`` docstring for more info on how metadata is used.

    Parameters
    ----------
    positive_scores: torch.tensor, 1-d
        Tensor containing scores for known positive items of shape
        ``num_negative_samples x batch_size``
    many_negative_scores: torch.tensor, 2-d
        Iterable of tensors containing scores for many (n > 1) sampled negative items of shape
        ``num_negative_samples x batch_size``. More tensors increase the likelihood of finding
        ranking-violating pairs, but risk overfitting
    num_items: int
        Total number of items in the dataset
    positive_items: torch.tensor, 1-d
        Tensor containing ids for known positive items of shape
        ``num_negative_samples x batch_size``. This is only needed if ``metadata`` is provided
    negative_items: torch.tensor, 2-d
        Tensor containing ids for sampled negative items of shape
        ``num_negative_samples x batch_size``. This is only needed if ``metadata`` is provided
    metadata: dict
        Keys should be strings identifying each metadata type that match keys in
        ``metadata_weights``. Values should be a ``torch.tensor`` of shape (num_items x 1). Each
        tensor should contain categorical metadata information about items (e.g. a number
        representing the genre of the item)
    metadata_weights: dict
        Keys should be strings identifying each metadata type that match keys in ``metadata``.
        Values should be the amount of weight to place on a match of that type of metadata, with the
        sum of all values ``<= 1``.
        e.g. If ``metadata_weights = {'genre': .3, 'director': .2}``, then an item is:

        * a 100% match if it's the same item,

        * a 50% match if it's a different item with the same genre and same director,

        * a 30% match if it's a different item with the same genre and different director,

        * a 20% match if it's a different item with a different genre and same director,

        * a 0% match if it's a different item with a different genre and different director,
          which is equivalent to the loss without any partial credit

    Returns
    -------
    loss: torch.tensor

    References
    ----------
    .. [4] Weston et al. WSABIE: Scaling Up To Large Vocabulary Image Annotation.
        www.thespermwhale.com/jaseweston/papers/wsabie-ijcai.pdf.

    """
    if negative_items is not None and positive_items is not None:
        positive_items = positive_items.repeat([many_negative_scores.shape[0], 1])

    if metadata is not None and len(metadata) > 0:
        ideal_difference = ideal_difference_from_metadata(
            positive_items=positive_items,
            negative_items=negative_items,
            metadata=metadata,
            metadata_weights=metadata_weights,
        ).transpose(1, 0)
    else:
        ideal_difference = 1

    # device to put new tensors on
    device = positive_scores.device

    # WARP loss requires a different structure for positive and negative samples
    positive_scores = positive_scores.view(len(positive_scores), 1)
    many_negative_scores = torch.transpose(many_negative_scores, 0, 1)

    batch_size, max_trials = many_negative_scores.size(0), many_negative_scores.size(1)

    flattened_new_row_indices = torch.arange(0, batch_size, 1).long().to(device) * (max_trials + 1)
    tensor_of_ones = torch.ones(batch_size, 1).float().to(device)

    # ``initial_loss`` is essentially just hinge loss for now
    hinge_loss = ideal_difference - positive_scores + many_negative_scores

    # Add column of ones to so we know when we have used all our attempts. This is used for indexing
    # and computing ``should_count_loss`` if no real value is above 0.
    initial_loss_with_ones = torch.cat([hinge_loss, tensor_of_ones], dim=1)
    # this will be modified in ``_find_first_loss_violation``
    initial_loss_with_ones_binary = torch.cat([hinge_loss, tensor_of_ones], dim=1)

    number_of_tries = _find_first_loss_violation(initial_loss_with_ones_binary, device)

    prediction_index_for_flattened_predictions = number_of_tries + flattened_new_row_indices

    number_of_tries = (number_of_tries + 1).float()

    # IMPORTANT CHANGE: normal WARP weighting has the numerator set to ``num_items - 1``, but we
    # have found this does not penalize when the last item in a negative item sequence ranks above a
    # positive item score. Adjusting the numerator as below penalizes this correctly. Additionally,
    # adding a floor function to the numerator can also have the same negative effect of not
    # not counting loss. See the original implementation as a comment below, and our modified,
    # harsher calculation implemented below.
    # loss_weights = torch.log(torch.floor((num_items - 1) / number_of_tries))
    loss_weights = torch.log((num_items / number_of_tries))

    # don't count loss if we used max number of attempts looking for a violation and didn't find one
    should_we_count_loss = (number_of_tries <= max_trials).float()

    loss = (
        loss_weights
        * (
            initial_loss_with_ones.flatten()[prediction_index_for_flattened_predictions]
        )
        * should_we_count_loss
    )

    return (loss.sum() + loss.pow(2).sum()) / len(positive_scores)


def _find_first_loss_violation(losses: torch.tensor,
                               device: Union[str, torch.device, torch.cuda.device]) -> torch.tensor:
    """
    Find the index of the first violation where ``1 - positive_score + negative_score`` is greater
    than 0.

    """
    # set all negative losses to 0 and all positive losses to 1
    losses[losses < 0] = 0
    losses[losses > 0] = 1

    # after this, maximum value will be the first non-zero (bad) loss
    reverse_indices = torch.arange(losses.shape[1], 0, -1).to(device)
    min_index_of_good_loss = losses * reverse_indices

    # report the first loss that is positive here (not 0-based indexed)
    number_of_tries = torch.argmax(min_index_of_good_loss, 1, keepdim=True).flatten()

    return number_of_tries



from functools import partial
from typing import Callable, Dict, Optional, Union

import torch
from torch import nn
import torch.nn.functional as F
from torch.optim.lr_scheduler import ReduceLROnPlateau

from collie.model.base import BasePipeline, INTERACTIONS_LIKE_INPUT, ScaledEmbedding, ZeroEmbedding
from collie.utils import get_init_arguments, merge_docstrings, trunc_normal


class DeepFM(BasePipeline):
    # NOTE: the full docstring is merged in with ``BasePipeline``'s using ``merge_docstrings``.
    # Only the description of new or changed parameters are included in this docstring
    """
    Training pipeline for a deep factorization model.

    ``DeepFM`` models combine a shallow factorization machine and a deep multilayer perceptron
    network in a single, unified model. The model consists of embedding tables for users and items,
    and model output is the sum of 1) factorization machine output of both embeddings (shallow) and
    2) MLP output for the concatenation of both embeddings (deep).

    The implementation here is meant to mimic its original implementation as specified here:
    https://arxiv.org/pdf/1703.04247.pdf [3]_

    All ``DeepFM`` instances are subclasses of the ``LightningModule`` class
    provided by PyTorch Lightning. This means to train a model, you will need a
    ``collie.model.CollieTrainer`` object, but the model can be saved and loaded without this
    ``Trainer`` instance. Example usage may look like:

    .. code-block:: python

        from collie.model import CollieTrainer, DeepFM


        model = DeepFM(train=train)
        trainer = CollieTrainer(model)
        trainer.fit(model)
        model.eval()

        # do evaluation as normal with ``model``

        model.save_model(filename='model.pth')
        new_model = DeepFM(load_model_path='model.pth')

        # do evaluation as normal with ``new_model``

    Parameters
    ----------
    embedding_dim: int
        Number of latent factors to use for the matrix factorization embedding table. For the MLP
        embedding table, the dimensionality will be calculated with the formula
        ``embedding_dim * (2 ** (num_layers - 1))``
    num_layers: int
        Number of MLP layers to apply. Each MLP layer will have its input dimension calculated with
        the formula ``embedding_dim * (2 ** (``num_layers`` - ``current_layer_number``))``
    final_layer: str or function
        Final layer activation function. Available string options include:

        * 'sigmoid'

        * 'relu'

        * 'leaky_relu'

    dropout_p: float
        Probability of dropout
    sparse: bool
        Whether or not to treat embeddings as sparse tensors. If ``True``, cannot use weight decay
        on the optimizer
    bias_lr: float
        Bias terms learning rate. If 'infer', will set equal to ``lr``
    optimizer: torch.optim or str
        If a string, one of the following supported optimizers:

        * ``'sgd'`` (for ``torch.optim.SGD``)

        * ``'adam'`` (for ``torch.optim.Adam``)

    bias_optimizer: torch.optim or str
        Optimizer for the bias terms. This supports the same string options as ``optimizer``, with
        the addition of ``infer``, which will set the optimizer equal to ``optimizer``. If
        ``bias_optimizer`` is ``None``, only a single optimizer will be created for all model
        parameters
    y_range: tuple
        Specify as ``(min, max)`` to apply a sigmoid layer to the output score of the model to get
        predicted ratings within the range of ``min`` and ``max``

    References
    ----------
    .. [3] Guo, Huifeng, et al. "DeepFM: A Factorization-Machine Based Neural Network for CTR
        Prediction." ArXiv.org, 13 Mar. 2017, arxiv.org/abs/1703.04247.

    """
    def __init__(self,
                 train: INTERACTIONS_LIKE_INPUT = None,
                 val: INTERACTIONS_LIKE_INPUT = None,
                 embedding_dim: int = 8,
                 num_layers: int = 3,
                 final_layer: Optional[Union[str, Callable]] = None,
                 dropout_p: float = 0.0,
                 lr: float = 1e-3,
                 bias_lr: Optional[Union[float, str]] = 1e-2,
                 lr_scheduler_func: Optional[Callable] = partial(ReduceLROnPlateau,
                                                                 patience=1,
                                                                 verbose=True),
                 weight_decay: float = 0.0,
                 optimizer: Union[str, Callable] = 'adam',
                 bias_optimizer: Optional[Union[str, Callable]] = 'sgd',
                 loss: Union[str, Callable] = 'hinge',
                 metadata_for_loss: Optional[Dict[str, torch.tensor]] = None,
                 metadata_for_loss_weights: Optional[Dict[str, float]] = None,
                 # y_range: Optional[Tuple[float, float]] = None,
                 load_model_path: Optional[str] = None,
                 map_location: Optional[str] = None):
        super().__init__(**get_init_arguments())

    __doc__ = merge_docstrings(BasePipeline, __doc__, __init__)

    def _setup_model(self, **kwargs) -> None:
        """
        Method for building model internals that rely on the data passed in.

        This method will be called after `prepare_data`.

        """
        self.user_embeddings = ScaledEmbedding(num_embeddings=self.hparams.num_users,
                                               embedding_dim=self.hparams.embedding_dim)
        self.item_embeddings = ScaledEmbedding(num_embeddings=self.hparams.num_items,
                                               embedding_dim=self.hparams.embedding_dim)
        self.user_biases = ZeroEmbedding(num_embeddings=self.hparams.num_users,
                                         embedding_dim=1)
        self.item_biases = ZeroEmbedding(num_embeddings=self.hparams.num_items,
                                         embedding_dim=1)
        self.user_global_bias = nn.Parameter(torch.zeros(1))
        self.item_global_bias = nn.Parameter(torch.zeros(1))

        mlp_modules = []
        input_size = self.hparams.embedding_dim * 2
        for i in range(self.hparams.num_layers):
            next_input_size = (
                int(
                    self.hparams.embedding_dim
                    * 2
                    * ((self.hparams.num_layers - i) / (self.hparams.num_layers + 1))
                )
            )
            mlp_modules.append(nn.Linear(input_size, next_input_size))
            mlp_modules.append(nn.ReLU())
            mlp_modules.append(nn.Dropout(p=self.hparams.dropout_p))
            input_size = next_input_size
        self.mlp_layers = nn.Sequential(*mlp_modules)

        self.predict_layer = nn.Linear(next_input_size, 1)

        for m in self.mlp_layers:
            if isinstance(m, nn.Linear):
                # initialization taken from the official repo:
                # https://github.com/hexiangnan/neural_collaborative_filtering/blob/master/NeuMF.py
                trunc_normal(m.weight.data, std=0.01)

        nn.init.kaiming_uniform_(self.predict_layer.weight, nonlinearity='relu')

        for m in self.modules():
            if isinstance(m, nn.Linear) and m.bias is not None:
                m.bias.data.zero_()

    def forward(self, users: torch.tensor, items: torch.tensor) -> torch.tensor:
        """
        Forward pass through the model.

        Parameters
        ----------
        users: tensor, 1-d
            Array of user indices
        items: tensor, 1-d
            Array of item indices

        Returns
        -------
        preds: tensor, 1-d
            Predicted ratings or rankings

        """
        user_embeddings = self.user_embeddings(users)
        item_embeddings = self.item_embeddings(items)

        # FM output
        embedding_sum = user_embeddings + item_embeddings
        embedding_squared_sum = torch.pow(user_embeddings, 2) + torch.pow(item_embeddings, 2)
        embeddings_difference = embedding_sum - embedding_squared_sum
        fm_output = torch.sum(embeddings_difference, dim=1)

        # MLP output
        concatenated_embeddings = torch.cat((user_embeddings, item_embeddings), -1)
        mlp_output = self.predict_layer(self.mlp_layers(concatenated_embeddings)).squeeze()

        prediction = fm_output + mlp_output

        if callable(self.hparams.final_layer):
            prediction = self.hparams.final_layer(prediction)
        elif self.hparams.final_layer == 'sigmoid':
            prediction = torch.sigmoid(prediction)
        elif self.hparams.final_layer == 'relu':
            prediction = F.relu(prediction)
        elif self.hparams.final_layer == 'leaky_relu':
            prediction = F.leaky_relu(prediction)
        elif self.hparams.final_layer is not None:
            raise ValueError(f'{self.hparams.final_layer} not valid final layer value!')

        return prediction.view(-1)

    def _get_item_embeddings(self) -> torch.tensor:
        """Get item embeddings on device."""
        return self.item_embeddings.weight.data



from typing import Dict, List

import torch


class ScaledEmbedding(torch.nn.Embedding):
    """Embedding layer that initializes its values to use a truncated normal distribution."""
    def reset_parameters(self) -> None:
        """Overriding default ``reset_parameters`` method."""
        self.weight.data.normal_(0, 1.0 / (self.embedding_dim * 2.5))


class ZeroEmbedding(torch.nn.Embedding):
    """Embedding layer with weights zeroed-out."""
    def reset_parameters(self) -> None:
        """Overriding default ``reset_parameters`` method."""
        self.weight.data.zero_()


class MultiOptimizer(object):
    """
    A simple class that allows us to wrap multiple optimizers into a single API typical of a single
    optimizer, with ``zero_grad`` and ``step`` methods.

    Parameters
    ----------
    optimizers: List of ``torch.optim.Optimizer``s

    """
    def __init__(self, optimizers: List[torch.optim.Optimizer]):
        assert isinstance(optimizers, list), f'Expected list, got {type(optimizers)}!'

        self.optimizers = optimizers

    def zero_grad(self) -> None:
        """Apply ``zero_grad`` to all optimizers."""
        for optimizer in self.optimizers:
            optimizer.zero_grad()


class MultiLRScheduler(object):
    """
    A simple class that allows us to wrap multiple learning rate schedulers into a single API
    typical of a single learning rate scheduler with a ``step`` method.

    Parameters
    ----------
    lr_schedulers: List of dictionaries
        Each dictionary must have key ``scheduler``, which contains the actual learning rate
        scheduler

    """
    def __init__(self, lr_schedulers: List[Dict[str, torch.optim.lr_scheduler._LRScheduler]]):
        assert isinstance(lr_schedulers, list), f'Expected list, got {type(lr_schedulers)}!'

        lr_schedulers_dicts_removed = [lr_scheduler['scheduler'] for lr_scheduler in lr_schedulers]

        self.lr_schedulers = lr_schedulers_dicts_removed

    def step(self, *args, **kwargs) -> None:
        """Apply ``step`` to all optimizers."""
        for lr_scheduler in self.lr_schedulers:
            lr_scheduler.step(*args, **kwargs)



import pytest
import torch


SCORES = torch.tensor([
    1.1, 1.3, 4.7, -7.234,
    -1.3, 0.7, 4.7, -2.2468,
    -4.7, 1.3, 3.56, -0.8924,
    4.01, 2.7, 3.7, -5.2468,
    3.89, 12.8, -1.7, -0.8143,
])


@pytest.fixture()
def positive_items():
    return torch.tensor([0, 1, 2, 3])


@pytest.fixture()
def negative_items():
    return torch.tensor([4, 5, 6, 7])


@pytest.fixture()
def many_negative_items():
    return torch.tensor([
        [4, 5, 6, 7],
        [8, 9, 10, 11],
        [12, 13, 14, 15],
        [16, 17, 18, 19],
    ])


@pytest.fixture()
def positive_scores(positive_items):
    return SCORES[positive_items]


@pytest.fixture()
def negative_scores(negative_items):
    return SCORES[negative_items]


@pytest.fixture()
def many_negative_scores(many_negative_items):
    return SCORES[many_negative_items]


@pytest.fixture()
def metadata_a():
    return torch.tensor([
        0, 1, 1, 2,
        0, 0, 0, 1,
        2, 1, 1, 2,
        0, 0, 1, 2,
        0, 0, 1, 2,
    ])


@pytest.fixture()
def metadata_a_diff():
    return torch.tensor([.8, 1, 1, 1])


@pytest.fixture()
def metadata_b():
    return torch.tensor([
        1, 2, 2, 3,
        3, 2, 1, 3,
        3, 2, 2, 2,
        1, 1, 2, 2,
        1, 1, 2, 2,
    ])


@pytest.fixture()
def metadata_a_and_2_diff():
    return torch.tensor([
        [.8, .7, 1, .7],
        [1, .5, .5, .8],
        [.5, 1, .5, .8],
        [.5, 1, .5, .8],
    ])



from abc import ABCMeta
from collections import OrderedDict
from functools import reduce
from typing import Callable, Dict, List, Optional, Tuple, Union

import torch

from collie.interactions import (ApproximateNegativeSamplingInteractionsDataLoader,
                                 Interactions,
                                 InteractionsDataLoader)
from collie.model.base import BasePipeline
from collie.utils import get_init_arguments, merge_docstrings


INTERACTIONS_LIKE_INPUT = Union[ApproximateNegativeSamplingInteractionsDataLoader,
                                Interactions,
                                InteractionsDataLoader]


class MultiStagePipeline(BasePipeline, metaclass=ABCMeta):
    """
    Multi-stage pipeline model architectures to inherit from.

    This model template is intended for models that train in distinct stages, with a different
    optimizer optimizing each step. This allows model components to be optimized with a set
    order in mind, rather than all at once, such as with the ``BasePipeline``.

    Generally, multi-stage models will have a training protocol like:

    .. code-block:: python

        from collie.model import CollieTrainer, SomeMultiStageModel


        model = SomeMultiStageModel(train=train)
        trainer = CollieTrainer(model)

        # fit stage 1
        trainer.fit(model)

        # fit stage 2
        trainer.max_epochs += 10
        model.advance_stage()
        trainer.fit(model)

        # fit stage 3
        trainer.max_epochs += 10
        model.advance_stage()
        trainer.fit(model)

        # ... and so on, until...

        model.eval()

    Just like with ``BasePipeline``, all subclasses MUST at least override the following methods:

    * ``_setup_model`` - Set up the model architecture

    * ``forward`` - Forward pass through a model

    For ``item_item_similarity`` to work properly, all subclasses are should also implement:

    * ``_get_item_embeddings`` - Returns item embeddings from the model

    Notes
    -----
    * With each call of ``trainer.fit``, the optimizer and learning rate scheduler state will reset.
    * When loading a multi-stage model in, the state will be set to the last possible state. This
      state may have a different ``forward`` calculation than other states.

    Parameters
    ----------
    optimizer_config_list: list of dict
        List of dictionaries containing the optimizer configurations for each stage's
        optimizer(s). Each dictionary must contain the following keys:

        * ``lr``: str
            Learning rate for the optimizer

        * ``optimizer``: ``torch.optim`` or ``str``

        * ``parameter_prefix_list``: List[str]
            List of string prefixes corressponding to the model components that should be
            optimized with this optimizer

        * ``stage``: str
            Name of stage

        This must be ordered with the intended progression of stages.

    """
    def __init__(self,
                 train: INTERACTIONS_LIKE_INPUT = None,
                 val: INTERACTIONS_LIKE_INPUT = None,
                 lr_scheduler_func: Optional[Callable] = None,
                 weight_decay: float = 0.0,
                 optimizer_config_list: List[Dict[str, Union[float, List[str], str]]] = None,
                 loss: Union[str, Callable] = 'hinge',
                 metadata_for_loss: Optional[Dict[str, torch.tensor]] = None,
                 metadata_for_loss_weights: Optional[Dict[str, float]] = None,
                 load_model_path: Optional[str] = None,
                 map_location: Optional[str] = None,
                 **kwargs):
        stage_list = None

        if load_model_path is None:
            if optimizer_config_list is None:
                raise ValueError(
                    'Must provide ``optimizer_config_list`` when initializing a new multi-stage '
                    'model!'
                )
            else:
                stage_list = list(
                    OrderedDict.fromkeys(
                        [optimizer_config['stage'] for optimizer_config in optimizer_config_list]
                    )
                )

        super().__init__(stage_list=stage_list,
                         **get_init_arguments())

        if load_model_path is None:
            # set stage if we have not already loaded it in and set it there
            self.hparams.stage = self.hparams.stage_list[0]
            self.set_stage(self.hparams.stage)

    __doc__ = merge_docstrings(BasePipeline, __doc__, __init__)

    def _load_model_init_helper(self, *args, **kwargs) -> None:
        super()._load_model_init_helper(*args, **kwargs)

        # set the stage to the last stage
        self.hparams.stage = self.hparams.stage_list[-1]
        print(f'Set ``self.hparams.stage`` to "{self.hparams.stage}"')

    def advance_stage(self) -> None:
        """Advance the stage to the next one in ``self.hparams.stage_list``."""
        stage = self.hparams.stage

        if stage in self.hparams.stage_list:
            stage_idx = self.hparams.stage_list.index(stage)
            if (stage_idx + 1) >= len(self.hparams.stage_list):
                raise ValueError(f'Cannot advance stage past {stage} - it is the final stage!')

            self.set_stage(stage=self.hparams.stage_list[stage_idx + 1])

    def set_stage(self, stage: str) -> None:
        """Set the model to the desired stage."""
        if stage in self.hparams.stage_list:
            self.hparams.stage = stage
            print(f'Set ``self.hparams.stage`` to "{self.hparams.stage}"')
        else:
            raise ValueError(
                f'{stage} is not a valid stage, please choose one of {self.hparams.stage_list}'
            )

    def _get_optimizer_parameters(
            self,
            optimizer_config: List[Dict[str, Union[float, List[str], str]]],
            include_weight_decay: bool = True,
            **kwargs
    ) -> List[Dict[str, Union[torch.tensor, float]]]:
        optimizer_parameters = [
            {
                'params': (
                    param for (name, param) in self.named_parameters()
                    if reduce(
                        lambda x, y: x or y,
                        [
                            name.startswith(prefix) for prefix in
                            optimizer_config['parameter_prefix_list']
                        ],
                        False,
                    )
                ),
                'lr': optimizer_config['lr'],
            }
        ]

        if include_weight_decay:
            weight_decay_dict = {'weight_decay': self.hparams.weight_decay}
            [d.update(weight_decay_dict) for d in optimizer_parameters]

        return optimizer_parameters

    def configure_optimizers(self) -> (
        Union[Tuple[List[Callable], List[Callable]], Tuple[Callable, Callable], Callable]
    ):
        """
        Configure optimizers and learning rate schedulers to use in optimization.

        This method will be called after `setup`.

        Creates an optimizer and learning rate scheduler for each configuration dictionary in
        ``self.hparams.optimizer_config_list``.

        """
        # since this is the only function that is called before each ``trainer.fit`` call, we will
        # also take this time to ensure any external data a model might rely on has been properly
        # moved to the device before training
        self._move_any_external_data_to_device()

        optimizer_config_list = [
            self._get_optimizer(self.optimizer, optimizer_config=optimizer_config)
            for optimizer_config in self.hparams.optimizer_config_list
        ]

        if self.lr_scheduler_func is not None:
            monitor = 'val_loss_epoch'
            if self.val_loader is None:
                monitor = 'train_loss_epoch'

            # add in optimizer to scheduler function
            scheduler_list = [
                {
                    'scheduler': self.lr_scheduler_func(optimizer_config),
                    'monitor': monitor,
                }
                for optimizer_config in optimizer_config_list
            ]

            return optimizer_config_list, scheduler_list

        else:
            return optimizer_config_list

    def optimizer_step(self,
                       epoch: int = None,
                       batch_idx: int = None,
                       optimizer: torch.optim.Optimizer = None,
                       optimizer_idx: int = None,
                       optimizer_closure: Optional[Callable] = None,
                       **kwargs) -> None:
        """
        Overriding Lightning's optimizer step function to only step the optimizer associated with
        the relevant stage.

        See here for more details:
        https://pytorch-lightning.readthedocs.io/en/stable/common/lightning_module.html#optimizer-step

        Parameters
        ----------
        epoch: int
            Current epoch
        batch_idx: int
            Index of current batch
        optimizer: torch.optim.Optimizer
            A PyTorch optimizer
        optimizer_idx: int
            If you used multiple optimizers, this indexes into that list
        optimizer_closure: Callable
            Closure for all optimizers

        """
        if self.hparams.optimizer_config_list[optimizer_idx]['stage'] == self.hparams.stage:
            optimizer.step(closure=optimizer_closure)



import os

import numpy as np
import pandas as pd
import pytest

from collie.interactions import ExplicitInteractions, HDF5Interactions, Interactions
from collie.utils import create_ratings_matrix, pandas_df_to_hdf5


@pytest.fixture()
def df_for_interactions():
    # this should exactly match ``ratings_matrix_for_interactions`` below
    return pd.DataFrame(data={
        'user_id': [0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 5],
        'item_id': [1, 2, 2, 3, 4, 5, 6, 7, 8, 9, 0, 3],
        'ratings': [1, 1, 2, 3, 4, 5, 1, 2, 3, 4, 5, 5],
    })


@pytest.fixture()
def ratings_matrix_for_interactions():
    # this should exactly match ``df_for_interactions`` above
    return np.array([[0, 1, 1, 0, 0, 0, 0, 0, 0, 0],
                     [0, 0, 2, 3, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 4, 5, 1, 0, 0, 0],
                     [0, 0, 0, 0, 0, 0, 0, 2, 3, 4],
                     [5, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 5, 0, 0, 0, 0, 0, 0]])


@pytest.fixture()
def sparse_ratings_matrix_for_interactions(df_for_interactions):
    return create_ratings_matrix(df=df_for_interactions,
                                 user_col='user_id',
                                 item_col='item_id',
                                 ratings_col='ratings',
                                 sparse=True)


@pytest.fixture()
def df_for_interactions_with_missing_ids():
    # we are missing item ID 7
    # this should exactly match ``ratings_matrix_for_interactions_with_missing_ids`` below
    return pd.DataFrame(data={
        'user_id': [0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 5],
        'item_id': [1, 2, 2, 3, 4, 5, 6, 0, 8, 9, 0, 3],
        'ratings': [1, 1, 2, 3, 4, 5, 1, 2, 3, 4, 5, 5],
    })


@pytest.fixture()
def ratings_matrix_for_interactions_with_missing_ids():
    # we are missing item ID 7
    # this should exactly match ``df_for_interactions_with_missing_ids`` above
    return np.array([[0, 1, 1, 0, 0, 0, 0, 0, 0, 0],
                     [0, 0, 2, 3, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 0, 4, 5, 1, 0, 0, 0],
                     [0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                     [5, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                     [0, 0, 0, 5, 0, 0, 0, 0, 0, 0]])


@pytest.fixture()
def sparse_ratings_matrix_for_interactions_with_missing_ids(df_for_interactions_with_missing_ids):
    return create_ratings_matrix(df=df_for_interactions_with_missing_ids,
                                 user_col='user_id',
                                 item_col='item_id',
                                 ratings_col='ratings',
                                 sparse=True)


@pytest.fixture()
def df_for_interactions_with_0_ratings():
    # ``df_for_interactions`` but with three extra interactions with ratings of 0
    return pd.DataFrame(data={
        'user_id': [0, 0, 1, 1, 2, 2, 3, 1, 2, 3, 3, 3, 4, 5, 5],
        'item_id': [1, 2, 2, 3, 4, 5, 2, 4, 6, 7, 8, 9, 0, 3, 4],
        'ratings': [1, 1, 2, 3, 4, 5, 0, 0, 1, 2, 3, 4, 5, 5, 0],
    })


@pytest.fixture()
def df_for_interactions_with_duplicates():
    # this should match ``df_for_interactions`` with duplicate user/item pairs added at the
    # following indices: ``0 & 1`` and ``12 & 13``
    return pd.DataFrame(data={
        'user_id': [0, 0, 0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 5, 5],
        'item_id': [1, 1, 2, 2, 3, 4, 5, 6, 7, 8, 9, 0, 3, 3],
        'ratings': [1, 2, 1, 2, 3, 4, 5, 1, 2, 3, 4, 5, 5, 4],
    })


@pytest.fixture()
def interactions_pandas(df_for_interactions):
    return Interactions(users=df_for_interactions['user_id'],
                        items=df_for_interactions['item_id'],
                        ratings=df_for_interactions['ratings'],
                        check_num_negative_samples_is_valid=False)


@pytest.fixture()
def interactions_matrix(ratings_matrix_for_interactions):
    return Interactions(mat=ratings_matrix_for_interactions,
                        check_num_negative_samples_is_valid=False)


@pytest.fixture()
def interactions_sparse_matrix(sparse_ratings_matrix_for_interactions):
    return Interactions(mat=sparse_ratings_matrix_for_interactions,
                        check_num_negative_samples_is_valid=False)


@pytest.fixture()
def explicit_interactions_pandas(df_for_interactions):
    return ExplicitInteractions(users=df_for_interactions['user_id'],
                                items=df_for_interactions['item_id'],
                                ratings=df_for_interactions['ratings'])


@pytest.fixture()
def explicit_interactions_matrix(ratings_matrix_for_interactions):
    return ExplicitInteractions(mat=ratings_matrix_for_interactions)


@pytest.fixture()
def explicit_interactions_sparse_matrix(sparse_ratings_matrix_for_interactions):
    return ExplicitInteractions(mat=sparse_ratings_matrix_for_interactions)


@pytest.fixture()
def hdf5_pandas_df_path(df_for_interactions, tmpdir):
    hdf5_path = os.path.join(str(tmpdir), 'df_for_interactions.h5')
    pandas_df_to_hdf5(df=df_for_interactions, out_path=hdf5_path, key='interactions')

    return hdf5_path


@pytest.fixture()
def hdf5_pandas_df_path_with_meta(df_for_interactions, tmpdir):
    hdf5_path = os.path.join(str(tmpdir), 'df_for_interactions_meta.h5')
    pandas_df_to_hdf5(df=df_for_interactions, out_path=hdf5_path, key='interactions')

    additional_info_df = pd.DataFrame({
        'num_users': [df_for_interactions['user_id'].max() + 1],
        'num_items': [df_for_interactions['item_id'].max() + 1],
    })
    pandas_df_to_hdf5(df=additional_info_df, out_path=hdf5_path, key='meta')

    return hdf5_path


@pytest.fixture(params=['users', 'items', 'both_users_and_items'])
def hdf5_pandas_df_path_ids_start_at_1(request, df_for_interactions, tmpdir):
    incremented_df_for_interactions = df_for_interactions

    if 'users' in request.param:
        incremented_df_for_interactions['user_id'] += 1
    if 'items' in request.param:
        incremented_df_for_interactions['item_id'] += 1

    hdf5_path = os.path.join(str(tmpdir), 'df_for_interactions_incremented.h5')
    pandas_df_to_hdf5(df=incremented_df_for_interactions, out_path=hdf5_path, key='interactions')

    return hdf5_path


@pytest.fixture()
def hdf5_interactions(hdf5_pandas_df_path_with_meta):
    return HDF5Interactions(hdf5_path=hdf5_pandas_df_path_with_meta,
                            user_col='user_id',
                            item_col='item_id')



from datetime import datetime
import inspect
from pathlib import Path
import re
import time
from typing import Any, Dict, Iterable, List, Optional, Union

import docstring_parser
import numpy as np
import pandas as pd
import pytorch_lightning
from scipy.sparse import coo_matrix
import torch

from collie.interactions import Interactions


# helpful constants just for clearer code
NEWLINE_CHARACTER = '\n'
FOUR_SPACES = '    '
NEWLINE_CHARACTER_FOUR_SPACES = NEWLINE_CHARACTER + FOUR_SPACES


def get_random_seed() -> int:
    """Generate a random seed based on the current datetime."""
    return int(datetime.now().replace(microsecond=0).timestamp())


def create_ratings_matrix(df: pd.DataFrame,
                          user_col: str = 'user_id',
                          item_col: str = 'item_id',
                          ratings_col: str = 'rating',
                          sparse: bool = False) -> (coo_matrix, np.array):
    """
    Helper function to convert a Pandas DataFrame to 2-dimensional matrix.

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe with columns for user IDs, item IDs, and ratings
    user_col: str
        Column name for the user IDs
    item_col: str
        Column name for the item IDs
    ratings_col: str
        Column name for the ratings column
    sparse: bool
        Whether to return data as a sparse ``coo_matrix`` (True) or np.array (False)

    Returns
    -------
    ratings_matrix: np.array or scipy.sparse.coo_matrix, 2-d
        Data with users as rows, items as columns, and ratings as values

    """
    if sparse:
        ratings_matrix = _create_sparse_ratings_matrix_helper(users=df[user_col],
                                                              items=df[item_col],
                                                              ratings=df[ratings_col])
    else:
        if df[user_col].min() != 0 or df[item_col].min() != 0:
            raise ValueError(
                'Minimum values of ``df[user_col]`` and ``df[item_col]`` must both be 0.'
            )

        ratings_df = df.pivot(index=user_col, columns=item_col, values=ratings_col).fillna(0)
        ratings_matrix = ratings_df.to_numpy()

    return ratings_matrix


def _create_sparse_ratings_matrix_helper(users: Iterable[int],
                                         items: Iterable[int],
                                         ratings: Optional[Iterable[int]] = None,
                                         num_users: Union[int, str] = 'infer',
                                         num_items: Union[int, str] = 'infer') -> coo_matrix:
    """Create a sparse matrix from a series of arrays."""
    num_users = _infer_num_if_needed_for_1d_array(num_users, users)
    num_items = _infer_num_if_needed_for_1d_array(num_items, items)

    if ratings is None:
        ratings = np.ones_like(users)

    return coo_matrix(
        (np.array(ratings), (np.array(users), np.array(items))), shape=(num_users, num_items)
    )


def _infer_num_if_needed_for_1d_array(num: Union[int, str], array: Iterable[int]) -> int:
    """Return ``num`` or, if ``None``, the maximum value of ``array`` + 1."""
    if num == 'infer':
        num = max(array) + 1

    return num


def df_to_interactions(df: pd.DataFrame,
                       user_col: str = 'user_id',
                       item_col: str = 'item_id',
                       ratings_col: Optional[str] = 'rating',
                       **kwargs) -> Interactions:
    """
    Helper function to convert a DataFrame to an ``Interactions`` object.

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe with columns for user IDs, item IDs, and (optionally) ratings
    user_col: str
        Column name for the user IDs
    item_col: str
        Column name for the item IDs
    ratings_col: str
        Column name for the ratings column. If ``None``, will default to ratings of all 1s
    **kwargs
        Keyword arguments to pass to ``Interactions``

    Returns
    -------
    interactions: collie.interactions.Interactions

    """
    ratings = df[ratings_col] if ratings_col is not None else None

    return Interactions(users=df[user_col], items=df[item_col], ratings=ratings, **kwargs)


def convert_to_implicit(explicit_df: pd.DataFrame,
                        min_rating_to_keep: Optional[float] = 4,
                        user_col: str = 'user_id',
                        item_col: str = 'item_id',
                        ratings_col: str = 'rating') -> pd.DataFrame:
    """
    Convert explicit interactions data to implicit data.

    Duplicate user ID and item ID pairs will be dropped, as well as all scores that are
    ``< min_rating_to_keep``. All remaining interactions will have a rating of ``1``.

    Parameters
    ----------
    explicit_df: pd.DataFrame
        Dataframe with explicit ratings in the rating column
    min_rating_to_keep: int
        Minimum rating to be considered a valid interaction
    ratings_col: str
        Column name for the ratings column

    Returns
    -------
    implicit_df: pd.DataFrame
        Dataframe that converts all ``ratings >= min_rating_to_keep`` to 1 and drops the rest with a
        reset index. Note that the order of ``implicit_df`` will not be equal to ``explicit_df``

    """
    implicit_df = explicit_df.copy()

    implicit_df = (
        implicit_df
        .sort_values(by=ratings_col)
        .drop_duplicates(subset=[user_col, item_col], keep='last', ignore_index=True)
    )
    implicit_df = implicit_df.drop(implicit_df[implicit_df[ratings_col] < min_rating_to_keep].index)
    implicit_df[ratings_col] = 1

    return implicit_df.reset_index(drop=True)


def remove_users_with_fewer_than_n_interactions(df: pd.DataFrame,
                                                min_num_of_interactions: int = 3,
                                                user_col: str = 'user_id') -> pd.DataFrame:
    """
    Remove DataFrame rows with users who appear fewer than ``min_num_of_interactions`` times.

    Parameters
    ----------
    df: pd.DataFrame
    min_num_of_interactions: int
        Minimum number of interactions a user can have while remaining in ``filtered_df``
    user_col: str
        Column name for the user IDs

    Returns
    -------
    filtered_df: pd.DataFrame

    """
    value_counts_df = df[user_col].value_counts()

    return (
        df[~df[user_col].isin(
            value_counts_df[value_counts_df < min_num_of_interactions].index
        )].reset_index(drop=True)
    )


def trunc_normal(embedding_weight: torch.tensor,
                 mean: float = 0.0,
                 std: float = 1.0) -> torch.tensor:
    """
    Truncated normal initialization (approximation).

    Taken from FastAI: https://github.com/fastai/fastai/blob/master/fastai/layers.py

    """
    # From https://discuss.pytorch.org/t/implementing-truncated-normal-initializer/4778/12
    return embedding_weight.normal_().fmod_(2).mul_(std).add_(mean)


def get_init_arguments(exclude: Optional[Iterable[str]] = [],
                       verbose: bool = False) -> Dict[str, Any]:
    """
    Get all input arguments (*args and **kwargs) sent to the most-recently called method, given it
    is an ``__init__`` of a class.

    Parameters
    ----------
    exclude: list
        Arguments to exclude from ``init_args``. If an argument is not found in ``init_args``, it
        will be ignored
    verbose: bool
        Print keys in ``exclude`` not found in ``init_args``

    Returns
    -------
    init_args: dict
        Argument dictionary with keys being argument names and values being arguments

    Note
    ----
    If the most-recently called method is not an ``__init__`` of a class, this function will return
    an empty dictionary.

    """
    frame = inspect.currentframe().f_back
    init_args = pytorch_lightning.utilities.parsing.get_init_args(frame)

    if exclude:
        for exclude_arg in exclude:
            try:
                del init_args[exclude_arg]
            except KeyError:
                if verbose:
                    print(f'Key {exclude_arg} not found in ``init_args`` and will be ignored.')
                continue

    return init_args


def pandas_df_to_hdf5(df: pd.DataFrame,
                      out_path: Union[str, Path],
                      key: str = 'interactions') -> None:
    """Append a Pandas DataFrame to HDF5 using a ``table`` format and ``blosc`` compression."""
    df.to_hdf(str(out_path),
              key=key,
              mode='a',
              append=True,
              format='table',
              complib='blosc')


def df_to_html(df: pd.DataFrame,
               image_cols: List[str] = [],
               hyperlink_cols: List[str] = [],
               html_tags: Dict[str, Union[str, List[str]]] = dict(),
               transpose: bool = False,
               image_width: Optional[int] = None,
               max_num_rows: int = 200,
               **kwargs) -> str:
    """
    Convert a Pandas DataFrame to HTML.

    Parameters
    ----------
    df: DataFrame
        DataFrame to convert to HTML
    image_cols: str or list
        Column names that contain image urls or file paths. Columns specified as images will make
        all other transformations to those columns be ignored. Local files will display correctly in
        Jupyter if specified using relative paths but not if specified using absolute paths (see
        https://github.com/jupyter/notebook/issues/3810).
    hyperlink_cols: str or list
        Column names that contain hyperlinks to open in a new tab
    html_tags: dictionary
        A transformation to be inserted directly into the HTML tag.

        Ex: ``{'col_name_1': 'strong'}`` becomes ``<strong>col_name_1</strong>``

        Ex: ``{'col_name_2': 'mark'}`` becomes ``<mark>col_name_2</mark>``

        Ex: ``{'col_name_3': 'h2'}`` becomes ``<h2>col_name_3</h2>``

        Ex: ``{'col_name_4': ['em', 'strong']}`` becomes ``<em><strong>col_name_4</strong></em>``

    transpose: bool
        Transpose the DataFrame before converting to HTML
    image_width: int
        Set image width for each image generated
    max_num_rows: int
        Maximum number of rows to display
    **kwargs: keyword arguments
        Additional arguments sent to ``pandas.DataFrame.to_html``, as listed in:
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.to_html.html

    Returns
    -------
    df_html: HTML
        DataFrame converted to a HTML string, ready for displaying

    Examples
    --------
    In a Jupyter notebook:

    .. code-block:: python

        from IPython.core.display import display, HTML
        import pandas as pd


        df = pd.DataFrame({
            'item': ['Beefy Fritos® Burrito'],
            'price': ['1.00'],
            'image_url': ['https://www.tacobell.com/images/22480_beefy_fritos_burrito_269x269.jpg'],
        })
        display(
            HTML(
                df_to_html(
                    df,
                    image_cols='image_url',
                    html_tags={'item': 'strong', 'price': 'em'},
                    image_width=200,
                )
            )
        )

    Note
    ----
    Converted table will have CSS class 'dataframe', unless otherwise specified.

    """
    def _wrap_cols_if_needed(cols: [str, List[str]]) -> List[str]:
        """Necessary for columns named with integers."""
        try:
            iter(cols)
        except TypeError:
            cols = [cols]
        if isinstance(cols, str):
            cols = [cols]

        return cols

    if max_num_rows is None or len(df) <= max_num_rows:
        df = df.copy()  # copy the dataframe so we don't edit the original!
    else:
        # explicit copy eliminates a warning we don't need
        df = df.head(max_num_rows).copy()

    image_cols = _wrap_cols_if_needed(image_cols)
    for image_col in image_cols:
        if image_col not in df.columns:
            raise ValueError('{} not a column in df!'.format(image_col))
        if not image_width:
            df[image_col] = df[image_col].map(lambda x: f'<img src="{x}">')
        else:
            df[image_col] = df[image_col].map(lambda x: f'<img src="{x}" width={image_width}>')

    hyperlink_cols = _wrap_cols_if_needed(hyperlink_cols)
    for hyperlink_col in hyperlink_cols:
        if hyperlink_col not in df.columns:
            raise ValueError('{} not a column in df!'.format(hyperlink_col))
        if hyperlink_col in image_cols:
            continue
        df[hyperlink_col] = (
            df[hyperlink_col].map(lambda x: f'<a target="_blank" href="{x}">{x}</a>')
        )

    for col, transformations in html_tags.items():
        if col not in df.columns:
            raise ValueError(f'{col} not a column in df!')
        if col in image_cols:
            continue

        if isinstance(transformations, str):
            transformations = [transformations]

        opening_tag = ''
        for extra in transformations:
            opening_tag += f'<{extra}>'

        closing_tag = ''
        for extra in transformations[::-1]:
            closing_tag += f'</{extra}>'

        df[col] = df[col].map(lambda x: f'{opening_tag}{x}{closing_tag}')

    max_colwidth = pd.get_option('display.max_colwidth')
    if pd.__version__ != '0':
        # this option is not backwards compatible with Pandas v1.0.0
        pd.set_option('display.max_colwidth', None)
    else:
        pd.set_option('display.max_colwidth', -1)

    if transpose:
        df = df.T
    df_html = df.to_html(escape=False, **kwargs)

    pd.set_option('display.max_colwidth', max_colwidth)

    return df_html


class Timer(object):
    """Class to manage timing different sections of a job."""
    def __init__(self):
        self.start_time = time.time()
        self.current_time = self.start_time

    def timecheck(self, message: str = 'Finished') -> float:
        """Get time since last timecheck."""
        tmp_time = time.time()
        elapsed_time = (tmp_time-self.current_time)/60.0
        print('{0} ({1:.2f} min)'.format(message, elapsed_time))
        self.current_time = tmp_time

        return elapsed_time

    def time_since_start(self, message: str = 'Total time') -> float:
        """Get time since timer was instantiated."""
        total_time = (time.time() - self.start_time) / 60.0
        print('{0}: {1:.2f} min'.format(message, total_time))

        return total_time


def merge_docstrings(parent_class, child_docstring, child_class__init__):
    """
    Merge docstrings for Collie models to reduce the amount of repeated, shared docstrings.

    This method notes the arguments of the ``child_class``'s ``__init__`` function and searches
    the docstrings of both the child and parent (in order) to construct the docstring for the child
    class.

    Specifically, the final docstring returned will be, in order:

    ```
    CHILD SHORT DESCRIPTION

    CHILD LONG DESCRIPTION

    Parameters
    ----------
    for each ``arg`` in CHILD ``__init__`` ARGUMENTS:
        CHILD ARGUMENT DOCSTRING (if it exists in child docstring), else PARENT ARGUMENT DOCSTRING
    ...

    POST ``Parameters`` CHILD DOCSTRING, SIGNIFIED BY A REPEATED HYPHEN SEPARATOR

    ```

    Notes
    -----
    * The docstring returned will be ordered with a description immediately followed by the
      ``Parameters`` section.
    * ``Returns``, ``Raises``, and ``Deprecated`` sections are currently not supported and will be
      filtered out in the returned docstring.
    * Additional sections will be returned following the ``Parameters`` section if they are noted
      with a line of `-` the length of the title. If not, it will be filtered out.

    """
    # get parent class documentation
    parent_docstring = parent_class.__doc__

    # find the line where the ``Parameters`` section begins
    child_docstring_list = child_docstring.split(NEWLINE_CHARACTER)

    child_parameters_idx = [
        idx for idx, arg in enumerate(child_docstring_list)
        if re.search('\\sParameters\\s?$', arg)
    ]

    if len(child_parameters_idx) == 0:
        # no ``Parameters`` section is bad, fail early
        return child_docstring

    # parse both parent and child docstrings
    parent_parse = docstring_parser.numpydoc.NumpydocParser().parse(parent_docstring)
    child_parse = docstring_parser.numpydoc.NumpydocParser().parse(child_docstring)

    parent_arg_name_idx_dict = {
        param.arg_name: idx for idx, param in enumerate(parent_parse.params)
    }
    child_arg_name_idx_dict = {
        param.arg_name: idx for idx, param in enumerate(child_parse.params)
    }

    # list all arguments the child class's ``__init__`` method defines
    child_class_inspect_result = inspect.getfullargspec(child_class__init__)
    child_class_has_args = child_class_inspect_result.varargs is not None
    child_class_has_kwargs = child_class_inspect_result.varkw is not None

    child_class__init__args = child_class_inspect_result.args

    if child_class_has_args:
        child_class__init__args.append('*' + child_class_inspect_result.varargs)
    if child_class_has_kwargs:
        child_class__init__args.append('**' + child_class_inspect_result.varkw)

    # format the description of the class prior to the ``Parameters`` section
    short_description = child_parse.short_description if child_parse.short_description else ''
    long_description = child_parse.long_description if child_parse.long_description else ''

    if short_description:
        short_description = (
            NEWLINE_CHARACTER_FOUR_SPACES
            + short_description.replace(NEWLINE_CHARACTER, NEWLINE_CHARACTER_FOUR_SPACES)
            + NEWLINE_CHARACTER
        )

    if long_description:
        long_description = (
            NEWLINE_CHARACTER_FOUR_SPACES
            + long_description.replace(NEWLINE_CHARACTER, NEWLINE_CHARACTER_FOUR_SPACES)
            + NEWLINE_CHARACTER * 2
        )
    else:
        long_description = NEWLINE_CHARACTER

    final_docstring = short_description + long_description

    # loop through each expected argument, check if the docstring exists in the child (preferred)
    # or the parent docuementation, then add that to the final docstring
    if len(child_class__init__args) > 0:
        final_docstring += (
            FOUR_SPACES
            + 'Parameters'
            + NEWLINE_CHARACTER_FOUR_SPACES
            + '----------'
            + NEWLINE_CHARACTER
        )

        for arg in child_class__init__args:
            if arg in child_arg_name_idx_dict:
                param_idx = child_arg_name_idx_dict[arg]
                param = child_parse.params[param_idx]
            elif arg in parent_arg_name_idx_dict:
                param_idx = parent_arg_name_idx_dict[arg]
                param = parent_parse.params[param_idx]
            else:
                # argument isn't in the docstring, we can skip it
                continue

            arg_name = param.arg_name if param.arg_name else ''
            type_name = ': ' + param.type_name if param.type_name else ''
            description = param.description if param.description else ''

            final_docstring += (
                f'{FOUR_SPACES}{arg_name.strip()}{type_name.strip()}'
                f'{NEWLINE_CHARACTER}'
            )

            if description != '':
                final_docstring += FOUR_SPACES + FOUR_SPACES + (
                    description
                    .strip()
                    .replace(NEWLINE_CHARACTER, NEWLINE_CHARACTER_FOUR_SPACES + FOUR_SPACES)
                ) + NEWLINE_CHARACTER

    # add in the rest of the docstring post-``Parameters`` section
    for x in child_parse.meta:
        if type(x) == docstring_parser.DocstringMeta:
            final_docstring += (
                NEWLINE_CHARACTER_FOUR_SPACES
                + x.args[0].title()
                + NEWLINE_CHARACTER_FOUR_SPACES
                + '-' * len(x.args[0])
                + NEWLINE_CHARACTER_FOUR_SPACES
            )

            if x.description != '':
                final_docstring += (
                    x.description
                    .strip()
                    .replace(NEWLINE_CHARACTER, NEWLINE_CHARACTER_FOUR_SPACES)
                ) + NEWLINE_CHARACTER

    # replace lines that are just spaces with a newline character only
    final_docstring = re.sub(r'\n(\s)*\n',
                             NEWLINE_CHARACTER + NEWLINE_CHARACTER,
                             final_docstring)

    final_docstring += NEWLINE_CHARACTER_FOUR_SPACES

    return final_docstring



from functools import partial
from typing import Callable, Dict, Optional, Tuple, Union

import torch
from torch import nn
from torch.optim.lr_scheduler import ReduceLROnPlateau

from collie.model.base import BasePipeline, INTERACTIONS_LIKE_INPUT, ScaledEmbedding, ZeroEmbedding
from collie.utils import get_init_arguments, merge_docstrings


class MLPMatrixFactorizationModel(BasePipeline):
    # NOTE: the full docstring is merged in with ``BasePipeline``'s using ``merge_docstrings``.
    # Only the description of new or changed parameters are included in this docstring
    """
    Training pipeline for the matrix factorization model with MLP layers instead of a final dot
    product (like in ``MatrixFactorizationModel``).

    ``MLPMatrixFactorizationModel`` models have an embedding layer for both users and items which,
    are concatenated and sent through a MLP to output a single float ranking value.

    All ``MLPMatrixFactorizationModel`` instances are subclasses of the ``LightningModule`` class
    provided by PyTorch Lightning. This means to train a model, you will need a
    ``collie.model.CollieTrainer`` object, but the model can be saved and loaded without this
    ``Trainer`` instance. Example usage may look like:

    .. code-block:: python

        from collie.model import CollieTrainer, MLPMatrixFactorizationModel


        model = MLPMatrixFactorizationModel(train=train)
        trainer = CollieTrainer(model)
        trainer.fit(model)
        model.eval()

        # do evaluation as normal with ``model``

        model.save_model(filename='model.pth')
        new_model = MLPMatrixFactorizationModel(load_model_path='model.pth')

        # do evaluation as normal with ``new_model``

    Parameters
    ----------
    embedding_dim: int
        Number of latent factors to use for user and item embeddings
    num_layers: int
        Number of MLP layers to apply. Each MLP layer will have its input dimension calculated with
        the formula ``embedding_dim * (2 ** (``num_layers`` - ``current_layer_number``))``
    dropout_p: float
        Probability of dropout on the linear layers
    bias_lr: float
        Bias terms learning rate. If 'infer', will set equal to ``lr``
    optimizer: torch.optim or str
        If a string, one of the following supported optimizers:

        * ``'sgd'`` (for ``torch.optim.SGD``)

        * ``'adam'`` (for ``torch.optim.Adam``)

    bias_optimizer: torch.optim or str
        Optimizer for the bias terms. This supports the same string options as ``optimizer``, with
        the addition of ``infer``, which will set the optimizer equal to ``optimizer``. If
        ``bias_optimizer`` is ``None``, only a single optimizer will be created for all model
        parameters
    y_range: tuple
        Specify as ``(min, max)`` to apply a sigmoid layer to the output score of the model to get
        predicted ratings within the range of ``min`` and ``max``

    """
    def __init__(self,
                 train: INTERACTIONS_LIKE_INPUT = None,
                 val: INTERACTIONS_LIKE_INPUT = None,
                 embedding_dim: int = 30,
                 num_layers: int = 3,
                 dropout_p: float = 0.0,
                 lr: float = 1e-3,
                 bias_lr: Optional[Union[float, str]] = 1e-2,
                 lr_scheduler_func: Optional[Callable] = partial(ReduceLROnPlateau,
                                                                 patience=1,
                                                                 verbose=True),
                 weight_decay: float = 0.0,
                 optimizer: Union[str, Callable] = 'adam',
                 bias_optimizer: Optional[Union[str, Callable]] = 'sgd',
                 loss: Union[str, Callable] = 'hinge',
                 metadata_for_loss: Optional[Dict[str, torch.tensor]] = None,
                 metadata_for_loss_weights: Optional[Dict[str, float]] = None,
                 y_range: Optional[Tuple[float, float]] = None,
                 load_model_path: Optional[str] = None,
                 map_location: Optional[str] = None):
        super().__init__(**get_init_arguments())

    __doc__ = merge_docstrings(BasePipeline, __doc__, __init__)

    def _setup_model(self, **kwargs) -> None:
        """
        Method for building model internals that rely on the data passed in.

        This method will be called after `prepare_data`.

        """
        self.user_biases = ZeroEmbedding(num_embeddings=self.hparams.num_users,
                                         embedding_dim=1)
        self.item_biases = ZeroEmbedding(num_embeddings=self.hparams.num_items,
                                         embedding_dim=1)
        self.user_embeddings = ScaledEmbedding(num_embeddings=self.hparams.num_users,
                                               embedding_dim=self.hparams.embedding_dim)
        self.item_embeddings = ScaledEmbedding(num_embeddings=self.hparams.num_items,
                                               embedding_dim=self.hparams.embedding_dim)

        mlp_modules = []
        input_size = self.hparams.embedding_dim * 2
        for i in range(self.hparams.num_layers):
            next_input_size = (
                int(
                    self.hparams.embedding_dim
                    * 2
                    * ((self.hparams.num_layers - i) / (self.hparams.num_layers + 1))
                )
            )
            mlp_modules.append(nn.Linear(input_size, next_input_size))
            mlp_modules.append(nn.ReLU())
            mlp_modules.append(nn.Dropout(p=self.hparams.dropout_p))
            input_size = next_input_size
        self.mlp_layers = nn.Sequential(*mlp_modules)

        self.predict_layer = nn.Linear(next_input_size, 1)

    def forward(self, users: torch.tensor, items: torch.tensor) -> torch.tensor:
        """
        Forward pass through the model, roughly:

        ```prediction = MLP(concatenate(user_embedding * item_embedding)) + user_bias + item_bias```

        If dropout is added, it is applied for the two embeddings and not the biases.

        Parameters
        ----------
        users: tensor, 1-d
            Array of user indices
        items: tensor, 1-d
            Array of item indices

        Returns
        -------
        preds: tensor, 1-d
            Predicted ratings or rankings

        """
        user_embeddings = self.user_embeddings(users)
        item_embeddings = self.item_embeddings(items)

        concatenated_embeddings = torch.cat((user_embeddings, item_embeddings), -1)
        mlp_output = torch.sigmoid(
            self.predict_layer(
                self.mlp_layers(concatenated_embeddings)
            )
        ).squeeze()

        preds = (
            mlp_output
            + self.user_biases(users).squeeze(1)
            + self.item_biases(items).squeeze(1)
        )

        if self.hparams.y_range is not None:
            preds = (
                torch.sigmoid(preds)
                * (self.hparams.y_range[1] - self.hparams.y_range[0])
                + self.hparams.y_range[0]
            )

        return preds

    def _get_item_embeddings(self) -> torch.tensor:
        """Get item embeddings on device."""
        return self.item_embeddings.weight.data



import numpy as np
import pandas as pd
import pytest
import torch


@pytest.fixture()
def explicit_df():
    return pd.DataFrame(data={
        'userId': [2, 0, 1, 0, 1, 0, 3],
        'itemId': [0, 1, 1, 0, 3, 2, 1],
        'rating': [1, 2, 3, 1, 5, 3, 4]
    })


@pytest.fixture()
def explicit_df_with_duplicate_user_item_pairs():
    # the first and last user-item pairs have been repeated with modified ratings.
    # other than that, this is exactly the same as ``explicit_df`` above
    return pd.DataFrame(data={
        'userId': [2, 2, 2, 0, 1, 0, 1, 0, 3, 3, 3],
        'itemId': [0, 0, 0, 1, 1, 0, 3, 2, 1, 1, 1],
        'rating': [1, 3, 5, 2, 3, 1, 5, 3, 1, 4, 2]
    })


@pytest.fixture()
def df_to_turn_to_interactions():
    return pd.DataFrame(data={
        'userId': [0, 3, 1, 2, 1, 2, 1, 4],
        'itemId': [2, 0, 2, 2, 1, 4, 3, 2],
    })


@pytest.fixture()
def df_with_users_interacting_only_once():
    return pd.DataFrame(data={
        'userId': [0, 1, 1, 2, 2, 2, 3, 3, 3, 4, 0, 5],
        'itemId': [1, 2, 3, 4, 5, 6, 7, 8, 9, 1, 2, 3],
    })


@pytest.fixture()
def item_embeddings():
    return torch.Tensor([[1, 2, 3],
                         [4, 5, 6],
                         [7, 8, 9]])


@pytest.fixture()
def item_biases():
    return np.array([[10.1],
                     [11],
                     [12]])


@pytest.fixture()
def iid_to_item_dict():
    return {0: 'item_1', 1: 'item_2', 2: 'item_3'}


@pytest.fixture()
def df_html_test():
    d = {'title': ['Greg',
                   'Real Greg'],
         'description': ['some text here',
                         'more text here'],
         'link': ['https://madeupsite.com',
                  'https://anothermadeupsite.com'],
         'image': ['https://avatars0.githubusercontent.com/u/13399445',
                   'https://avatars3.githubusercontent.com/u/31417712']}

    df_html_test = pd.DataFrame(data=d)
    df_html_test = df_html_test[['title', 'description', 'link', 'image']]

    return df_html_test



import math
import random
from typing import List, Optional, Tuple

import numpy as np
import torch

from collie.interactions.datasets import HDF5Interactions, Interactions


class ApproximateNegativeSampler(torch.utils.data.sampler.Sampler):
    """
    Custom ``Sampler`` for bulk-sampling approximate negative items in ``Interactions`` data.

    Parameters
    ----------
    interactions: Interactions
    batch_size: int
        Number of samples per batch to load
    shuffle: bool
        Whether to shuffle the order of batches returned or not. This is especially useful for
        training data to ensure the model does not overfit to a specific order of data
    seed: int
        Seed for shuffling if ``shuffle is True``

    """
    def __init__(self,
                 interactions: Interactions,
                 batch_size: int = 1024,
                 shuffle: bool = False,
                 seed: Optional[int] = None):
        self.interactions = interactions
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.seed = seed

        self.iteration_order = np.arange(len(self.interactions))

        np.random.seed(self.seed)

    def __iter__(self) -> 'ApproximateNegativeSampler':
        """Setup iteration through ``ApproximateNegativeSamplingInteractionsDataLoader`` data."""
        if self.shuffle:
            np.random.shuffle(self.iteration_order)

        # reset pointer
        self._pointer = 0

        return self

    def __next__(self) -> np.array:
        """Get the indices for the next batch of data."""
        if self._pointer >= len(self.interactions):
            raise StopIteration

        idxs = self.iteration_order[self._pointer:(self._pointer + self.batch_size)]

        self._pointer += self.batch_size

        return idxs

    def __len__(self) -> int:
        """Number of batches returned in the sampler."""
        return math.ceil(len(self.interactions) / self.batch_size)


class HDF5Sampler(torch.utils.data.sampler.Sampler):
    """
    Custom ``Sampler`` for HDF5 data, with each sampled item being a start index and a batch size
    to use in ``HDF5Interactions.__getitem__``.

    Parameters
    ----------
    hdf5_interactions: HDF5Interactions
    batch_size: int
        Number of samples per batch to load
    shuffle: bool
        Whether to shuffle the order of batches returned or not. This is especially useful for
        training data to ensure the model does not overfit to a specific order of data. Note that
        this will not perform a true shuffle of the data, but will instead shuffle the order of
        batches. While this is an approximation of true sampling, it allows us a greater speed up
        during model training for a negligible effect on model performance
    seed: int
        Seed for shuffling if ``shuffle is True``

    """
    def __init__(self,
                 hdf5_interactions: HDF5Interactions,
                 batch_size: int = 1024,
                 shuffle: bool = False,
                 seed: Optional[int] = None):
        self.hdf5_interactions = hdf5_interactions
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.seed = seed

        self.data_to_iterate_through = [
            (start_idx, self.batch_size)
            for start_idx in range(0, len(self.hdf5_interactions), self.batch_size)
        ]

        random.seed(self.seed)

    def __iter__(self) -> 'HDF5Sampler':
        """Setup iteration through ``HDF5Sampler`` data."""
        if self.shuffle:
            random.shuffle(self.data_to_iterate_through)

        # reset pointer
        self._pointer = 0

        return self

    def __next__(self) -> List[Tuple[int]]:
        """Get the indices for the next batch of data."""
        if self._pointer >= len(self.data_to_iterate_through):
            raise StopIteration

        idx = self.data_to_iterate_through[self._pointer]

        self._pointer += 1

        return idx

    def __len__(self) -> int:
        """Number of batches returned in the sampler."""
        return math.ceil(len(self.hdf5_interactions) / self.batch_size)
